import os
import asyncio
import hashlib
import pathlib

import synapse.tests.utils as s_t_utils
import synapse.tools.pullfile as s_pullfile

class TestPullFile(s_t_utils.SynTest):

    async def test_pullfile(self):

        async with self.getTestAxon() as axon:

            axonurl = axon.getLocalUrl()

            testhash = hashlib.sha256(b'test').hexdigest()
            visihash = hashlib.sha256(b'visi').hexdigest()
            nonehash = hashlib.sha256(b'none').hexdigest()

            testbash = hashlib.sha256(b'test').digest()
            visibash = hashlib.sha256(b'visi').digest()

            self.eq(((4, visibash), (4, testbash)), await axon.puts([b'visi', b'test']))

            with self.getTestDir() as wdir:

                outp = self.getTestOutp()
                self.eq(0, await s_pullfile.main(['-a', axonurl,
                                            '-o', wdir,
                                            '-l', testhash,
                                            '-l', nonehash], outp))
                oldcwd = os.getcwd()
                os.chdir(wdir)
                self.eq(0, await s_pullfile.main(['-a', axonurl,
                                            '-l', visihash], outp))

                os.chdir(oldcwd)

                with open(pathlib.Path(wdir, testhash), 'rb') as fd:
                    self.eq(b'test', fd.read())

                with open(pathlib.Path(wdir, visihash), 'rb') as fd:
                    self.eq(b'visi', fd.read())

                self.true(outp.expect(f'{nonehash} not in axon store'))
                self.true(outp.expect(f'Fetching {testhash} to file'))
                self.true(outp.expect(f'Fetching {visihash} to file'))



import synapse.exc as s_exc
import synapse.lib.dyndeps as s_dyndeps
import synapse.tests.utils as s_t_utils

hehe = 'woot'

class Foo:
    def bar(self):
        return 'baz'

def woot(x, y=30):
    return x + y

class DynDepsTest(s_t_utils.SynTest):

    def test_dyndeps_dynmod(self):
        with self.getLoggerStream('synapse.lib.dyndeps', 'Failed to import "- -"') as stream:
            self.none(s_dyndeps.getDynMod('- -'))
            self.true(stream.wait(1))
        self.nn(s_dyndeps.getDynMod('sys'))

    def test_dyndeps_dynloc(self):
        self.none(s_dyndeps.getDynLocal('synapse.tests.test_lib_dyndeps.gronk'))
        self.nn(s_dyndeps.getDynLocal('synapse.tests.test_lib_dyndeps.hehe'))

    def test_dyndeps_dyntask(self):
        task = ('synapse.tests.test_lib_dyndeps.Foo', (), {})
        foo = s_dyndeps.runDynTask(task)
        self.eq(foo.bar(), 'baz')

    def test_dyndeps_nosuchdyn(self):
        self.raises(s_exc.NoSuchDyn, s_dyndeps.tryDynMod, 'newpnewp')
        self.raises(s_exc.NoSuchDyn, s_dyndeps.tryDynLocal, 'sys.newpnewp')

    def test_dyndeps_meth(self):
        self.nn(s_dyndeps.getDynMeth('synapse.lib.base.Base.fini'))
        self.none(s_dyndeps.getDynMeth('synapse.lib.base.Base.newp'))



import argparse

import synapse.exc as s_exc
import synapse.lib.output as s_output

class Parser(argparse.ArgumentParser):

    def __init__(self, prog=None, outp=s_output.stdout, **kwargs):

        self.outp = outp
        self.exited = False

        argparse.ArgumentParser.__init__(self,
                                         prog=prog,
                                         formatter_class=argparse.RawDescriptionHelpFormatter,
                                         **kwargs)

    def exit(self, status=0, message=None):
        '''
        Argparse expects exit() to be a terminal function and not return.
        As such, this function must raise an exception instead.
        '''
        self.exited = True
        self.status = status

        if message is not None:
            self.outp.printf(message)
        raise s_exc.ParserExit(mesg=message, status=status)

    def _print_message(self, text, fd=None):
        '''
        Note:  this overrides an existing method in ArgumentParser
        '''
        self.outp.printf(text)



import shutil
import asyncio

import synapse.exc as s_exc
import synapse.common as s_common

import synapse.lib.coro as s_coro
import synapse.lib.lmdbslab as s_lmdbslab
import synapse.lib.multislabseqn as s_multislabseqn

import synapse.tests.utils as s_t_utils
from synapse.tests.utils import alist

class MultiSlabSeqn(s_t_utils.SynTest):

    async def test_multislabseqn(self):

        with self.getTestDir() as dirn:

            async with await s_multislabseqn.MultiSlabSeqn.anit(dirn) as msqn:
                s_multislabseqn.logger.debug(f'Repr test {msqn}')

                self.eq(0, msqn.index())

                retn = await alist(msqn.iter(0))
                self.eq([], retn)

                retn = await msqn.last()
                self.eq(None, retn)

                retn = await msqn.rotate()
                self.eq(0, retn)

                self.false(await msqn.cull(0))

                retn = await msqn.add('foo')
                self.eq(0, retn)

                retn = await msqn.add('foo2')
                self.eq(1, retn)

                retn = await alist(msqn.iter(0))
                self.eq([(0, 'foo'), (1, 'foo2')], retn)

                retn = await alist(msqn.iter(1))
                self.eq([(1, 'foo2')], retn)

                retn = await msqn.add('foo9', indx=9)
                self.eq(9, retn)

                retn = await alist(msqn.iter(0))
                self.eq([(0, 'foo'), (1, 'foo2'), (9, 'foo9')], retn)

                evnt1 = msqn.getOffsetEvent(9)
                self.true(evnt1.is_set())
                self.true(await msqn.waitForOffset(9, timeout=0.5))

                evnt1 = msqn.getOffsetEvent(10)
                self.false(await msqn.waitForOffset(10, timeout=0.1))

                retn = await msqn.add('foo10')
                self.eq(10, retn)

                retn = await alist(msqn.iter(0))
                self.eq([(0, 'foo'), (1, 'foo2'), (9, 'foo9'), (10, 'foo10')], retn)

                retn = await msqn.last()
                self.eq((10, 'foo10'), retn)

                retn = msqn.index()
                self.eq(11, retn)

                self.eq(4, msqn.tailseqn.size)
                fns = sorted(s_common.listdir(dirn, glob='*.lmdb'))
                self.len(1, fns)

                # cause a rotation
                self.eq(11, await msqn.rotate())
                self.eq(11, msqn.index())
                self.eq((10, 'foo10'), await msqn.last())

                retn = await alist(msqn.iter(0))
                self.eq([(0, 'foo'), (1, 'foo2'), (9, 'foo9'), (10, 'foo10')], retn)

                self.len(0, [x for x in msqn.tailseqn.iter(0)])

                fns = sorted(s_common.listdir(dirn, glob='*.lmdb'))
                self.len(2, fns)

                # Need one entry so can't cull at >= 10
                self.false(await msqn.cull(10))
                self.false(await msqn.cull(11))

                self.true(await msqn.cull(9))
                fns = sorted(s_common.listdir(dirn, glob='*.lmdb'))
                self.len(2, fns)

                retn = await alist(msqn.iter(0))
                self.eq([(10, 'foo10')], retn)

                # Once we write into tailseqn we can actually remove the rotated seqn
                await msqn.add('foo11')
                retn = await alist(msqn.iter(0))
                self.eq([(10, 'foo10'), (11, 'foo11')], retn)

                self.true(await msqn.cull(10))
                fns = sorted(s_common.listdir(dirn, glob='*.lmdb'))
                self.len(1, fns)

                retn = await alist(msqn.iter(0))
                self.eq([(11, 'foo11')], retn)
                self.eq((11, 'foo11'), await msqn.last())

                # Add some values and rotate for persistence check
                await msqn.add('foo12')
                await msqn.rotate()
                await msqn.add('foo13')
                await msqn.add('foo14')

                exp = [(11, 'foo11'), (12, 'foo12'), (13, 'foo13'), (14, 'foo14')]
                self.eq(exp, await alist(msqn.iter(0)))

            # Persistence check

            async with await s_multislabseqn.MultiSlabSeqn.anit(dirn) as msqn:
                self.eq(exp, await alist(msqn.iter(0)))

                self.eq((14, 'foo14'), await msqn.last())

                self.eq('foo11', await msqn.get(11))
                self.eq('foo12', await msqn.get(12))
                self.eq('foo13', await msqn.get(13))
                self.eq('foo14', await msqn.get(14))

                retn = await alist(msqn.gets(9, wait=False))
                self.eq(exp, retn)

                evnt = asyncio.Event()

                async def getter():
                    retn = []
                    async for item in msqn.gets(9):
                        evnt.set()
                        if item[1] == 'done':
                            return retn
                        retn.append(item)
                    return retn

                task = msqn.schedCoro(getter())
                await s_coro.event_wait(evnt, timeout=1)
                await msqn.add('done')

                retn = await asyncio.wait_for(task, timeout=1)
                self.eq(exp, retn)

                # Add entries not on the tail
                retn = await msqn.add('foo11b', indx=11)
                self.eq(11, retn)

                retn = await msqn.add('foo13b', indx=13)
                self.eq(13, retn)

                await self.asyncraises(s_exc.BadIndxValu, msqn.add('foo7', indx=7))

                retn = await alist(msqn.iter(1))
                exp = [(11, 'foo11b'), (12, 'foo12'), (13, 'foo13b'), (14, 'foo14'), (15, 'done')]
                self.eq(exp, retn)

                # Give a chance for the non-iterated async generators to get cleaned up
                await asyncio.sleep(0)
                await asyncio.sleep(0)

                # Make sure we're not holding onto more than 2 slabs

                # rotate
                await msqn.add('foo16')
                await msqn.rotate()
                await msqn.add('foo17')
                await msqn.add('foo18')

                fns = sorted(s_common.listdir(dirn, glob='*.lmdb'))
                self.len(3, fns)

                self.len(2, msqn._openslabs)

                retn = await msqn.get(11)  # first
                self.eq('foo11b', retn)

                self.len(2, msqn._openslabs)

                retn = await msqn.get(14)  # middle
                self.eq('foo14', retn)

                self.len(2, msqn._openslabs)

                retn = await msqn.get(17)  # tail
                self.eq('foo17', retn)

                self.len(2, msqn._openslabs)

                # Make sure we don't open the same slab twice

                # Keep a ref to the first slab
                it = msqn.iter(0)
                retn = await it.__anext__()
                self.eq(retn, (11, 'foo11b'))

                self.len(2, msqn._openslabs)

                # (Need to evict first slab ref from the cacheslab)
                self.true(msqn._cacheslab.path.endswith('b.lmdb'))
                retn = await msqn.get(14)
                self.eq(retn, 'foo14')
                self.true(msqn._cacheslab.path.endswith('d.lmdb'))

                # Should have the tail slab, the cache slab, and the open iterator slabs
                self.len(3, msqn._openslabs)

                retn = await msqn.get(12)
                self.eq(retn, 'foo12')

                retn = await alist(it)
                self.len(7, retn)

                # Iterator exhausted: should have just the cache slab (10) and the tail slab (20)
                self.len(2, msqn._openslabs)

    async def test_multislabseqn_cull(self):

        with self.getTestDir() as dirn:

            async with await s_multislabseqn.MultiSlabSeqn.anit(dirn) as msqn:

                for i in range(10):
                    await msqn.add(f'foo{i}')

                # cull with only one seqn
                self.true(await msqn.cull(2))
                await self.asyncraises(s_exc.BadIndxValu, msqn.get(2))

                await msqn.rotate()

                for i in range(10, 15):
                    await msqn.add(f'foo{i}')

                # A no-op
                self.false(await msqn.cull(99))

                # Ensure there's a cached slab
                retn = await msqn.get(4)
                self.eq('foo4', retn)

                self.true(await msqn.cull(4))
                await self.asyncraises(s_exc.BadIndxValu, msqn.get(4))
                await self.asyncraises(s_exc.BadIndxValu, msqn.add('foo12', indx=4))

                retn = await msqn.get(5)
                self.eq('foo5', retn)

                it = msqn.iter(0)
                retn = await it.__anext__()
                await self.asyncraises(s_exc.SlabInUse, msqn.cull(6))
                await it.aclose()

                # culling on the tail just moves firstindx forward
                it = msqn.iter(10)
                retn = await it.__anext__()
                await msqn.cull(11)
                self.eq(12, msqn.firstindx)
                await it.aclose()

                retn = await alist(msqn.iter(1))
                self.eq([(12, 'foo12'), (13, 'foo13'), (14, 'foo14')], retn)

                await self.asyncraises(s_exc.BadIndxValu, msqn.get(11))
                await self.asyncraises(s_exc.BadIndxValu, msqn.add('foo', indx=11))

                # Make sure it deleted a slab
                fns = sorted(s_common.listdir(dirn, glob='*.lmdb'))
                self.len(1, fns)

                # Make sure ranges are updated
                self.eq([10], msqn._ranges)

                # Can't cull before firstidx
                self.false(await msqn.cull(10))

            async with await s_multislabseqn.MultiSlabSeqn.anit(dirn) as msqn:
                await msqn.cull(13)
                retn = await alist(msqn.iter(1))
                self.eq([(14, 'foo14')], retn)

    async def test_multislabseqn_discover(self):
        '''
        Test all the horrible things that admins can do by deleting/moving slabs
        '''

        # Speed up copying dirs
        slabopts = {'map_size': 100000}

        with self.getTestDir() as dirn:

            origdirn = s_common.gendir(dirn, 'orig')

            async with await s_multislabseqn.MultiSlabSeqn.anit(origdirn, slabopts=slabopts) as msqn:
                for i in range(25):
                    if i > 0 and i % 10 == 0:
                        await msqn.rotate()
                    await msqn.add(f'foo{i}')

            baddirn = s_common.genpath(dirn, 'bad1')
            shutil.copytree(origdirn, baddirn)

            # Make a slab a non-dir
            slab0dirn = s_common.genpath(baddirn, f'seqn{"0" * 16}.lmdb')
            shutil.rmtree(slab0dirn)
            s_common.jssave('{}', slab0dirn)

            with self.getAsyncLoggerStream('synapse.lib.multislabseqn', 'non-directory') as stream:
                async with await s_multislabseqn.MultiSlabSeqn.anit(baddirn) as msqn:
                    await self.agenlen(15, msqn.iter(0))
                await stream.wait(timeout=1)

            # Switcheroo

            baddirn = s_common.genpath(dirn, 'bad2')
            shutil.copytree(origdirn, baddirn)

            slab0dirn = s_common.genpath(baddirn, f'seqn{"0" * 16}.lmdb')
            slab10dirn = s_common.genpath(baddirn, f'seqn{"0" * 14}0a.lmdb')
            tmpdirn = s_common.genpath(baddirn, 'tmp')
            shutil.move(slab10dirn, tmpdirn)
            shutil.move(slab0dirn, slab10dirn)
            shutil.move(tmpdirn, slab0dirn)

            with self.raises(s_exc.BadCoreStore):
                async with await s_multislabseqn.MultiSlabSeqn.anit(baddirn) as msqn:
                    pass

            # Delete out from the middle
            baddirn = s_common.genpath(dirn, 'bad3')
            shutil.copytree(origdirn, baddirn)

            slab10dirn = s_common.genpath(baddirn, f'seqn{"0" * 14}0a.lmdb')
            shutil.rmtree(slab10dirn)

            with self.getAsyncLoggerStream('synapse.lib.multislabseqn', 'gap in indices') as stream:
                async with await s_multislabseqn.MultiSlabSeqn.anit(baddirn) as msqn:
                    await self.agenlen(15, msqn.iter(0))
                await stream.wait(timeout=1)

            # Wipe a seqn clean
            baddirn = s_common.genpath(dirn, 'bad4')
            shutil.copytree(origdirn, baddirn)

            slab20dirn = s_common.genpath(baddirn, f'seqn{"0" * 14}14.lmdb')
            async with await s_lmdbslab.Slab.anit(slab20dirn) as slab:
                seqn = slab.getSeqn('nexuslog')
                await seqn.cull(25)

            async with await s_multislabseqn.MultiSlabSeqn.anit(baddirn) as msqn:
                await self.agenlen(20, msqn.iter(0))

            # Overlapping seqns
            baddirn = s_common.genpath(dirn, 'bad5')
            shutil.copytree(origdirn, baddirn)

            slab10dirn = s_common.genpath(baddirn, f'seqn{"0" * 14}0a.lmdb')
            async with await s_lmdbslab.Slab.anit(slab10dirn) as slab:
                seqn = slab.getSeqn('nexuslog')
                seqn.add('foo', indx=22)

            with self.raises(s_exc.BadCoreStore):
                async with await s_multislabseqn.MultiSlabSeqn.anit(baddirn) as msqn:
                    pass

            # Somebody really messing with us
            baddirn = s_common.genpath(dirn, 'bad6')
            shutil.copytree(origdirn, baddirn)

            slab20dirn = s_common.genpath(baddirn, f'seqn{"0" * 14}14.lmdb')
            async with await s_lmdbslab.Slab.anit(slab20dirn) as slab:
                db = slab.initdb('info')
                slab.put(b'firstindx', s_common.int64en(99), db=db)

            with self.raises(s_exc.BadCoreStore):
                async with await s_multislabseqn.MultiSlabSeqn.anit(baddirn) as msqn:
                    pass

    async def test_multislabseqn_iterrotate(self):

        with self.getTestDir() as dirn:

            async with await s_multislabseqn.MultiSlabSeqn.anit(dirn) as msqn:

                for i in range(9):
                    await msqn.add(f'foo{i}')

                it = msqn.iter(0)
                retn = await it.__anext__()
                self.eq((0, 'foo0'), retn)

                # rotate
                await msqn.add('foo9')
                await msqn.rotate()
                await msqn.add('foo10')
                await msqn.add('foo11')

                self.len(2, msqn._openslabs)

                # can still iterate on the seqn
                retn = await it.__anext__()
                self.eq((1, 'foo1'), retn)

                # iterator will pick up the new seqn
                retn = await alist(it)
                self.len(10, retn)
                self.eq([(2, 'foo2'), (11, 'foo11')], [retn[0], retn[-1]])
                await it.aclose()

                # and we can cull once done
                await msqn.cull(10)
                self.len(1, msqn._openslabs)

                self.eq(12, msqn.tailseqn.indx)

    async def test_multislabseqn_multicache(self):
        '''
        A new consumer can iterate over an older non-cacheslab while the cacheslab is in use
        '''

        with self.getTestDir() as dirn:

            async with await s_multislabseqn.MultiSlabSeqn.anit(dirn) as msqn:

                for i in range(15):
                    if i > 0 and i % 4 == 0:
                        await msqn.rotate()
                    await msqn.add(f'foo{i}')

                self.len(4, msqn._ranges)
                self.none(msqn._cacheslab)
                self.len(1, msqn._openslabs)

                it00 = msqn.iter(5)
                retn = await it00.__anext__()
                self.eq((5, 'foo5'), retn)

                self.nn(msqn._cacheslab)
                self.true(msqn._cacheslab.path.endswith('4.lmdb'))
                self.len(2, msqn._openslabs)

                it01 = msqn.iter(0)
                retn = await it01.__anext__()
                self.eq((0, 'foo0'), retn)

                self.true(msqn._cacheslab.path.endswith('0.lmdb'))
                self.len(3, msqn._openslabs)

                retn = await it00.__anext__()
                self.eq((6, 'foo6'), retn)

                retn = await it01.__anext__()
                self.eq((1, 'foo1'), retn)

                await it00.aclose()
                await it01.aclose()

                self.true(msqn._cacheslab.path.endswith('0.lmdb'))
                self.len(2, msqn._openslabs)

                # cache slab remains open, but can be culled
                await msqn.cull(13)
                self.none(msqn._cacheslab)
                self.len(1, msqn._openslabs)

    async def test_multislabseqn_last(self):

        with self.getTestDir() as dirn:

            async with await s_multislabseqn.MultiSlabSeqn.anit(dirn) as msqn:

                self.none(await msqn.last())

                for i in range(5):
                    await msqn.add(f'foo{i}')

                self.eq((4, 'foo4'), await msqn.last())

                # rotate so we have an empty tail slab
                await msqn.rotate()
                self.eq(0, msqn.tailseqn.size)

                self.eq((4, 'foo4'), await msqn.last())

                # create a hole in the index
                await msqn.add('foo6', indx=6)
                self.eq((6, 'foo6'), await msqn.last())



import os
import json

import synapse.common as s_common

import synapse.lib.cell as s_cell
import synapse.lib.jupyter as s_jupyter
import synapse.lib.msgpack as s_msgpack
import synapse.lib.stormsvc as s_stormsvc

import synapse.tests.utils as s_t_utils

class TstsvcApi(s_cell.CellApi, s_stormsvc.StormSvc):
    _storm_svc_name = 'testsvc'
    _storm_svc_pkgs = (
        {
            'name': 'testsvc',
            'version': (0, 0, 1),
            'synapse_minversion': (2, 8, 0),
            'commands': (
                {
                    'name': 'testsvc.magic',
                    'storm': '[inet:ipv4=0]',
                },
            )
        },
    )

    async def testmeth(self):
        return 'shazam'

class Tstsvc(s_cell.Cell):
    cellapi = TstsvcApi

class JupyterTest(s_t_utils.SynTest):
    testmods = ['synapse.tests.utils.TestModule']

    async def test_tempcorecmdr(self):
        outp = self.getTestOutp()
        cmdrcore = await s_jupyter.getTempCoreCmdr(self.testmods, outp)
        self.false(cmdrcore.isfini)
        nodes = await cmdrcore.eval('[test:str=beep]', cmdr=True)
        self.len(1, nodes)
        self.eq(nodes[0][0], ('test:str', 'beep'))
        self.true(outp.expect('cli> storm [test:str=beep]'))
        await cmdrcore.fini()
        self.true(cmdrcore.isfini)

    async def test_tempcoreprox(self):
        prox = await s_jupyter.getTempCoreProx(self.testmods)
        self.false(prox.isfini)
        nodes = await s_t_utils.alist(prox.eval('[test:str=beep]'))
        self.len(1, nodes)
        self.eq(nodes[0][0], ('test:str', 'beep'))
        await prox.fini()
        self.true(prox.isfini)

    async def test_tempcorecmdrstormsvc(self):
        cmdrcore, svcprox = await s_jupyter.getTempCoreCmdrStormsvc('testsvc', Tstsvc.anit)

        self.false(cmdrcore.isfini)
        self.false(svcprox.isfini)

        mesgs = await cmdrcore.storm('service.list')
        self.true(any(['True (testsvc)' in str(mesg) for mesg in mesgs]))

        nodes = await cmdrcore.eval('testsvc.magic')
        self.len(1, nodes)

        self.eq('shazam', await svcprox.testmeth())

        await cmdrcore.fini()
        self.true(cmdrcore.isfini)

        await svcprox.fini()
        self.true(svcprox.isfini)

    async def test_cmdrcore(self):

        async with self.getTestCoreAndProxy() as (realcore, core):

            outp = self.getTestOutp()
            async with await s_jupyter.CmdrCore.anit(core, outp=outp) as cmdrcore:
                podes = await cmdrcore.eval('[test:str=beep]',
                                            num=1, cmdr=False)
                self.len(1, podes)
                self.false(outp.expect('[test:str=beep]', throw=False))

                mesgs = await cmdrcore.storm('[test:str=boop]',
                                             num=1, cmdr=True)
                self.true(outp.expect('[test:str=boop]', throw=False))
                podes = [m[1] for m in mesgs if m[0] == 'node']
                self.gt(len(mesgs), len(podes))
                self.len(1, podes)
                self.eq(podes[0][0], ('test:str', 'boop'))

                # Opts works for cmdr=False
                podes = await cmdrcore.eval('[test:str=$foo]',
                                            {'vars': {'foo': 'duck'}},
                                            num=1, cmdr=False)
                self.len(1, podes)
                self.eq(podes[0][0], ('test:str', 'duck'))

                # Opts does not work with cmdr=True - we have no way to plumb it through.
                with self.getAsyncLoggerStream('synapse.lib.view',
                                               'Error during storm execution') as stream:
                    with self.raises(AssertionError):
                        await cmdrcore.eval('[test:str=$foo]',
                                              opts={'vars': {'foo': 'fowl'}},
                                              cmdr=True)
                    self.true(await stream.wait(6))

                # Assertion based tests
                podes = await cmdrcore.eval('test:int', num=0)
                self.len(0, podes)
                podes = await cmdrcore.eval('test:str', num=3)
                self.len(3, podes)
                await self.asyncraises(AssertionError, cmdrcore.eval('test:str', num=1))

                # Feed function for data loading
                data = [
                    (('test:int', 137), {}),
                ]

                await cmdrcore.addFeedData('syn.nodes', data)
                podes = await cmdrcore.eval('test:int=137',
                                            num=1, cmdr=False)
                self.len(1, podes)

        # Raw cmdline test
        async with self.getTestCoreAndProxy() as (realcore, core):

            outp = self.getTestOutp()
            async with await s_jupyter.CmdrCore.anit(core, outp=outp) as cmdrcore:
                await cmdrcore.runCmdLine('help')
                self.true(outp.expect('cli> help'))
                self.true(outp.expect('List commands and display help output.'))

    async def test_log_supression(self):

        async with self.getTestCoreAndProxy() as (realcore, core):

            outp = self.getTestOutp()
            async with await s_jupyter.CmdrCore.anit(core, outp=outp) as cmdrcore:
                with self.getAsyncLoggerStream('synapse.lib.view') as stream:
                    mesgs = await cmdrcore.storm('[test:int=beep]',
                                                 num=0, cmdr=False,
                                                 suppress_logging=True)
                    self.stormIsInErr('invalid literal for int', mesgs)
                stream.seek(0)
                self.notin('Error during storm execution', stream.read())

                with self.getAsyncLoggerStream('synapse.lib.view',
                                                     'Error during storm execution') as stream:
                    mesgs = await cmdrcore.storm('[test:int=beep]',
                                                 num=0, cmdr=False,
                                                 suppress_logging=False)
                    self.true(await stream.wait(6))
                    self.stormIsInErr('invalid literal for int', mesgs)

    def test_doc_data(self):
        with self.getTestDir() as dirn:
            s_common.gendir(dirn, 'docdata', 'stuff')

            docdata = s_common.genpath(dirn, 'docdata')

            root = s_common.genpath(dirn, 'synapse', 'userguides')

            d = {'key': 'value'}

            s_common.jssave(d, docdata, 'data.json')
            s_common.yamlsave(d, docdata, 'data.yaml')
            s_msgpack.dumpfile(d, os.path.join(docdata, 'data.mpk'))
            with s_common.genfile(docdata, 'stuff', 'data.txt') as fd:
                fd.write('beep'.encode())
            with s_common.genfile(docdata, 'data.jsonl') as fd:
                fd.write(json.dumps(d).encode() + b'\n')
                fd.write(json.dumps(d).encode() + b'\n')
                fd.write(json.dumps(d).encode() + b'\n')

            data = s_jupyter.getDocData('data.json', root)
            self.eq(data, d)
            data = s_jupyter.getDocData('data.yaml', root)
            self.eq(data, d)
            data = s_jupyter.getDocData('data.mpk', root)
            self.eq(data, d)
            data = s_jupyter.getDocData('stuff/data.txt', root)
            self.eq(data, b'beep')
            data = s_jupyter.getDocData('data.jsonl', root)
            self.eq(data, [d, d, d])

            self.raises(ValueError, s_jupyter.getDocData, 'newp.bin', root)
            self.raises(ValueError, s_jupyter.getDocData,
                        '../../../../../../etc/passwd', root)



import synapse.vendor.utils as s_v_utils

from synapse.vendor.xrpl.tests.test_main_test_cases import test_cases
from synapse.vendor.xrpl.core import addresscodec
from synapse.vendor.xrpl.core.addresscodec.main import MAX_32_BIT_UNSIGNED_INT


class TestMain(s_v_utils.VendorTest):
    def test_classic_address_to_xaddress(self):
        for test_case in test_cases:
            (
                classic_address,
                tag,
                expected_main_xaddress,
                expected_test_xaddress,
            ) = test_case

            # test
            xaddress = addresscodec.classic_address_to_xaddress(
                classic_address, tag, True
            )
            self.assertEqual(xaddress, expected_test_xaddress)

            # main
            xaddress = addresscodec.classic_address_to_xaddress(
                classic_address, tag, False
            )
            self.assertEqual(xaddress, expected_main_xaddress)

    def test_xaddress_to_classic_address(self):
        for test_case in test_cases:
            (
                expected_classic_address,
                expected_tag,
                main_xaddress,
                test_xaddress,
            ) = test_case

            # test
            classic_address, tag, is_test = addresscodec.xaddress_to_classic_address(
                test_xaddress
            )
            self.assertEqual(classic_address, expected_classic_address)
            self.assertEqual(tag, expected_tag)
            self.assertTrue(is_test)

            # main
            classic_address, tag, is_test = addresscodec.xaddress_to_classic_address(
                main_xaddress
            )
            self.assertEqual(classic_address, expected_classic_address)
            self.assertEqual(tag, expected_tag)
            self.assertFalse(is_test)

    def test_classic_address_to_xaddress_invalid_tag(self):
        classic_address = "rGWrZyQqhTp9Xu7G5Pkayo7bXjH4k4QYpf"
        tag = MAX_32_BIT_UNSIGNED_INT + 1

        self.assertRaises(
            addresscodec.XRPLAddressCodecException,
            addresscodec.classic_address_to_xaddress,
            classic_address,
            tag,
            True,
        )

        self.assertRaises(
            addresscodec.XRPLAddressCodecException,
            addresscodec.classic_address_to_xaddress,
            classic_address,
            tag,
            False,
        )

    def test_classic_address_to_xaddress_bad_classic_address(self):
        classic_address = "r"

        self.assertRaises(
            ValueError,
            addresscodec.classic_address_to_xaddress,
            classic_address,
            None,
            True,
        )

        self.assertRaises(
            ValueError,
            addresscodec.classic_address_to_xaddress,
            classic_address,
            None,
            False,
        )

    def test_is_valid_classic_address_secp256k1(self):
        classic_address = "rU6K7V3Po4snVhBBaU29sesqs2qTQJWDw1"

        result = addresscodec.is_valid_classic_address(classic_address)
        self.assertTrue(result)

    def test_is_valid_classic_address_ed25519(self):
        classic_address = "rLUEXYuLiQptky37CqLcm9USQpPiz5rkpD"

        result = addresscodec.is_valid_classic_address(classic_address)
        self.assertTrue(result)

    def test_is_valid_classic_address_invalid(self):
        classic_address = "rU6K7V3Po4snVhBBaU29sesqs2qTQJWDw2"

        result = addresscodec.is_valid_classic_address(classic_address)
        self.assertFalse(result)

    def test_is_valid_classic_address_empty(self):
        classic_address = ""

        result = addresscodec.is_valid_classic_address(classic_address)
        self.assertFalse(result)

    def test_is_valid_xaddress_valid(self):
        xaddress = "X7AcgcsBL6XDcUb289X4mJ8djcdyKaB5hJDWMArnXr61cqZ"

        result = addresscodec.is_valid_xaddress(xaddress)
        self.assertTrue(result)

    def test_is_valid_xaddress_invalid(self):
        xaddress = "XVLhHMPHU98es4dbozjVtdWzVrDjtV18pX8zeUygYrCgrPh"

        result = addresscodec.is_valid_xaddress(xaddress)
        self.assertFalse(result)

    def test_is_valid_xaddress_empty(self):
        xaddress = ""

        result = addresscodec.is_valid_xaddress(xaddress)
        self.assertFalse(result)



import logging

import synapse.exc as s_exc

import synapse.lib.types as s_types
import synapse.lib.module as s_module

import synapse.lookup.phonenum as s_l_phone

logger = logging.getLogger(__name__)


def digits(text):
    return ''.join([c for c in text if c.isdigit()])

def chop_imei(imei):
    valu = int(imei)
    tac = int(imei[0:8])
    snr = int(imei[8:14])
    cd = int(imei[14:15])
    return valu, {'subs': {'tac': tac, 'serial': snr, 'cd': cd}}

class Phone(s_types.Str):

    def postTypeInit(self):
        s_types.Str.postTypeInit(self)
        self.opts['globsuffix'] = True
        self.setNormFunc(str, self._normPyStr)
        self.setNormFunc(int, self._normPyInt)

    def _normPyStr(self, valu):
        digs = digits(valu)
        if not digs:
            raise s_exc.BadTypeValu(valu=valu, name=self.name,
                                    mesg='requires a digit string')
        subs = {}
        try:
            info = s_l_phone.getPhoneInfo(int(digs))
        except Exception as e:  # pragma: no cover
            raise s_exc.BadTypeValu(valu=valu, name=self.name,
                                    mesg='Failed to get phone info') from None
        cc = info.get('cc')
        if cc is not None:
            subs['loc'] = cc
        # TODO prefix based validation?
        return digs, {'subs': subs}

    def _normPyInt(self, valu):
        if valu < 1:
            raise s_exc.BadTypeValu(valu=valu, name=self.name,
                                    mesg='phone int must be greater than 0')
        return self._normPyStr(str(valu))

    def repr(self, valu):
        # XXX geo-aware reprs are practically a function of cc which
        # XXX the raw value may only have after doing a s_l_phone lookup
        if valu[0] == '1' and len(valu) == 11:
            area = valu[1:4]
            pref = valu[4:7]
            numb = valu[7:11]
            return '+1 (%s) %s-%s' % (area, pref, numb)

        return '+' + valu

def imeicsum(text):
    '''
    Calculate the imei check byte.
    '''
    digs = []
    for i in range(14):

        v = int(text[i])
        if i % 2:
            v *= 2

        [digs.append(int(x)) for x in str(v)]

    chek = 0
    valu = sum(digs)
    remd = valu % 10
    if remd != 0:
        chek = 10 - remd

    return str(chek)

class Imsi(s_types.Int):

    def postTypeInit(self):
        self.opts['size'] = 8
        self.opts['signed'] = False
        return s_types.Int.postTypeInit(self)

    def _normPyInt(self, valu):
        imsi = str(valu)
        ilen = len(imsi)
        if ilen > 15:
            raise s_exc.BadTypeValu(valu=valu, name=self.name,
                                    mesg='invalid imsi len: %d' % (ilen,))

        mcc = imsi[0:3]
        # TODO full imsi analysis tree
        return valu, {'subs': {'mcc': mcc}}

# TODO: support pre 2004 "old" imei format
class Imei(s_types.Int):

    def postTypeInit(self):
        self.opts['size'] = 8
        self.opts['signed'] = False
        return s_types.Int.postTypeInit(self)

    def _normPyInt(self, valu):
        imei = str(valu)
        ilen = len(imei)

        # we are missing our optional check digit
        # lets add it for consistency...
        if ilen == 14:
            imei += imeicsum(imei)
            return chop_imei(imei)

        # if we *have* our check digit, lets check it
        elif ilen == 15:
            if imeicsum(imei) != imei[-1]:
                raise s_exc.BadTypeValu(valu=valu, name=self.name,
                                        mesg='invalid imei checksum byte')
            return chop_imei(imei)

        raise s_exc.BadTypeValu(valu=valu, name=self.name,
                                mesg='Failed to norm IMEI')

class TelcoModule(s_module.CoreModule):
    def getModelDefs(self):
        modl = {
            'ctors': (

                ('tel:mob:imei', 'synapse.models.telco.Imei', {}, {
                    'ex': '490154203237518',
                    'doc': 'An International Mobile Equipment Id.'}),

                ('tel:mob:imsi', 'synapse.models.telco.Imsi', {}, {
                    'ex': '310150123456789',
                    'doc': 'An International Mobile Subscriber Id.'}),

                ('tel:phone', 'synapse.models.telco.Phone', {}, {
                    'ex': '+15558675309',
                    'doc': 'A phone number.'}),

            ),

            'types': (

                ('tel:call', ('guid', {}), {
                    'doc': 'A guid for a telephone call record.'}),

                ('tel:txtmesg', ('guid', {}), {
                    'doc': 'A guid for an individual text message.'}),

                ('tel:mob:tac', ('int', {}), {
                    'ex': '49015420',
                    'doc': 'A mobile Type Allocation Code.'}),

                ('tel:mob:imid', ('comp', {'fields': (('imei', 'tel:mob:imei'), ('imsi', 'tel:mob:imsi'))}), {
                    'ex': '(490154203237518, 310150123456789)',
                    'doc': 'Fused knowledge of an IMEI/IMSI used together.'}),

                ('tel:mob:imsiphone', ('comp', {'fields': (('imsi', 'tel:mob:imsi'), ('phone', 'tel:phone'))}), {
                    'ex': '(310150123456789, "+7(495) 124-59-83")',
                    'doc': 'Fused knowledge of an IMSI assigned phone number.'}),

                ('tel:mob:telem', ('guid', {}), {
                    'doc': 'A single mobile telemetry measurement.'}),

                ('tel:mob:mcc', ('str', {'regex': '^[0-9]{3}$', 'strip': 1}), {
                    'doc': 'ITU Mobile Country Code.',
                }),

                ('tel:mob:mnc', ('str', {'regex': '^[0-9]{2,3}$', 'strip': 1}), {
                    'doc': 'ITU Mobile Network Code.',
                }),

                ('tel:mob:carrier', ('comp', {'fields': (('mcc', 'tel:mob:mcc'), ('mnc', 'tel:mob:mnc'))}), {
                    'doc': 'The fusion of a MCC/MNC.'
                }),

                ('tel:mob:cell', ('comp', {'fields': (('carrier', 'tel:mob:carrier'),
                                                      ('lac', ('int', {})),
                                                      ('cid', ('int', {})))}), {
                    'doc': 'A mobile cell site which a phone may connect to.'
                }),

            ),

            'forms': (
                ('tel:phone', {}, (
                    ('loc', ('loc', {}), {
                        'doc': 'The location associated with the number.',
                    }),
                )),
                ('tel:call', {}, (
                    ('src', ('tel:phone', {}), {
                        'doc': 'The source phone number for a call.'
                    }),
                    ('dst', ('tel:phone', {}), {
                        'doc': 'The destination phone number for a call.'
                    }),
                    ('time', ('time', {}), {
                        'doc': 'The time the call was initiated.'
                    }),
                    ('duration', ('int', {}), {
                        'doc': 'The duration of the call in seconds.'
                    }),
                    ('connected', ('bool', {}), {
                        'doc': 'Indicator of whether the call was connected.',
                    }),
                    ('text', ('str', {}), {
                        'doc': 'The text transcription of the call.',
                        'disp': {'hint': 'text'},
                    }),
                    ('file', ('file:bytes', {}), {
                        'doc': 'A file containing related media.',
                    }),
                )),
                ('tel:txtmesg', {}, (
                    ('from', ('tel:phone', {}), {
                        'doc': 'The phone number assigned to the sender.'
                    }),
                    ('to', ('tel:phone', {}), {
                        'doc': 'The phone number assigned to the primary recipient.'
                    }),
                    ('recipients', ('array', {'type': 'tel:phone', 'uniq': True, 'sorted': True}), {
                        'doc': 'An array of phone numbers for additional recipients of the message.',
                    }),
                    ('svctype', ('str', {'enums': 'sms,mms,rcs', 'strip': 1, 'lower': 1}), {
                        'doc': 'The message service type (sms, mms, rcs).',
                    }),
                    ('time', ('time', {}), {
                        'doc': 'The time the message was sent.'
                    }),
                    ('text', ('str', {}), {
                        'doc': 'The text of the message.',
                        'disp': {'hint': 'text'},
                    }),
                    ('file', ('file:bytes', {}), {
                        'doc': 'A file containing related media.',
                    }),
                )),
                ('tel:mob:tac', {}, (
                    ('org', ('ou:org', {}), {
                        'doc': 'The org guid for the manufacturer.',
                    }),
                    ('manu', ('str', {'lower': 1}), {
                        'doc': 'The TAC manufacturer name.',
                    }),
                    ('model', ('str', {'lower': 1}), {
                        'doc': 'The TAC model name.',
                    }),
                    ('internal', ('str', {'lower': 1}), {
                        'doc': 'The TAC internal model name.',
                    }),
                )),
                ('tel:mob:imei', {}, (
                    ('tac', ('tel:mob:tac', {}), {
                        'ro': True,
                        'doc': 'The Type Allocate Code within the IMEI.'
                    }),
                    ('serial', ('int', {}), {
                        'ro': True,
                        'doc': 'The serial number within the IMEI.',
                    })
                )),
                ('tel:mob:imsi', {}, (
                    ('mcc', ('tel:mob:mcc', {}), {
                        'ro': True,
                        'doc': 'The Mobile Country Code.',
                    }),
                )),
                ('tel:mob:imid', {}, (
                    ('imei', ('tel:mob:imei', {}), {
                        'ro': True,
                        'doc': 'The IMEI for the phone hardware.'
                    }),
                    ('imsi', ('tel:mob:imsi', {}), {
                        'ro': True,
                        'doc': 'The IMSI for the phone subscriber.'
                    }),
                )),
                ('tel:mob:imsiphone', {}, (
                    ('phone', ('tel:phone', {}), {
                        'ro': True,
                        'doc': 'The phone number assigned to the IMSI.'
                    }),
                    ('imsi', ('tel:mob:imsi', {}), {
                        'ro': True,
                        'doc': 'The IMSI with the assigned phone number.'
                    }),
                )),
                ('tel:mob:mcc', {}, (
                    ('loc', ('loc', {}), {'doc': 'Location assigned to the MCC.'}),
                )),
                ('tel:mob:carrier', {}, (
                    ('mcc', ('tel:mob:mcc', {}), {
                        'ro': True,
                    }),
                    ('mnc', ('tel:mob:mnc', {}), {
                        'ro': True,
                    }),
                    ('org', ('ou:org', {}), {
                        'doc': 'Organization operating the carrier.'
                    }),
                    ('loc', ('loc', {}), {
                        'doc': 'Location the carrier operates from.'
                    }),
                )),
                ('tel:mob:cell', {}, (
                    ('carrier', ('tel:mob:carrier', {}), {'doc': 'Mobile carrier.', 'ro': True, }),
                    ('carrier:mcc', ('tel:mob:mcc', {}), {'doc': 'Mobile Country Code.', 'ro': True, }),
                    ('carrier:mnc', ('tel:mob:mnc', {}), {'doc': 'Mobile Network Code.', 'ro': True, }),
                    ('lac', ('int', {}), {'doc': 'Location Area Code. LTE networks may call this a TAC.',
                                          'ro': True, }),
                    ('cid', ('int', {}), {'doc': 'The Cell ID.', 'ro': True, }),
                    ('radio', ('str', {'lower': 1, 'onespace': 1}), {'doc': 'Cell radio type.'}),
                    ('latlong', ('geo:latlong', {}), {'doc': 'Last known location of the cell site.'}),

                    ('loc', ('loc', {}), {
                        'doc': 'Location at which the cell is operated.'}),

                    ('place', ('geo:place', {}), {
                        'doc': 'The place associated with the latlong property.'}),
                )),

                ('tel:mob:telem', {}, (

                    ('time', ('time', {}), {}),
                    ('latlong', ('geo:latlong', {}), {}),

                    ('http:request', ('inet:http:request', {}), {
                        'doc': 'The HTTP request that the telemetry was extracted from.',
                    }),

                    ('host', ('it:host', {}), {
                        'doc': 'The host that generated the mobile telemetry data.'}),

                    ('place', ('geo:place', {}), {
                        'doc': 'The place representing the location of the mobile telemetry sample.'}),

                    ('loc', ('loc', {}), {
                        'doc': 'The geo-political location of the mobile telemetry sample.',
                    }),

                    ('accuracy', ('geo:dist', {}), {
                        'doc': 'The reported accuracy of the latlong telemetry reading.',
                    }),

                    # telco specific data
                    ('cell', ('tel:mob:cell', {}), {}),
                    ('cell:carrier', ('tel:mob:carrier', {}), {}),
                    ('imsi', ('tel:mob:imsi', {}), {}),
                    ('imei', ('tel:mob:imei', {}), {}),
                    ('phone', ('tel:phone', {}), {}),

                    # inet protocol addresses
                    ('mac', ('inet:mac', {}), {}),
                    ('ipv4', ('inet:ipv4', {}), {}),
                    ('ipv6', ('inet:ipv6', {}), {}),

                    ('wifi', ('inet:wifi:ap', {}), {}),
                    ('wifi:ssid', ('inet:wifi:ssid', {}), {}),
                    ('wifi:bssid', ('inet:mac', {}), {}),

                    # host specific data
                    ('adid', ('it:adid', {}), {}),
                    ('aaid', ('it:os:android:aaid', {}), {}),
                    ('idfa', ('it:os:ios:idfa', {}), {}),

                    # User related data
                    ('name', ('ps:name', {}), {}),
                    ('email', ('inet:email', {}), {}),
                    ('acct', ('inet:web:acct', {}), {}),

                    # reporting related data
                    ('app', ('it:prod:softver', {}), {}),

                    ('data', ('data', {}), {}),
                    # any other fields may be refs...
                )),

            )
        }
        name = 'tel'
        return ((name, modl),)



import regex

import synapse.exc as s_exc

# TODO:  commonize with storm.lark
re_scmd = '[a-z][a-z0-9.]+'
scmdre = regex.compile(re_scmd)
univrestr = r'\.[a-z_][a-z0-9_]*([:.][a-z0-9_]+)*'
univre = regex.compile(univrestr)
proprestr = r'[a-z_][a-z0-9_]*(:[a-z0-9_]+)+([:.][a-z_ ][a-z0-9_]+)*'
proporunivrestr = f'({univrestr})|({proprestr})'
proporunivre = regex.compile(proporunivrestr)
propre = regex.compile(proprestr)
formrestr = r'[a-z_][a-z0-9_]*(:[a-z0-9_]+)+'
formre = regex.compile(formrestr)
tagrestr = r'(\w+\.)*\w+'
tagre = regex.compile(tagrestr)
basepropnopivpropstr = r'[a-z_][a-z0-9_]*(?:(\:|\.)[a-z_][a-z0-9_]*)*'
basepropnopivpropre = regex.compile(basepropnopivpropstr)

whites = set(' \t\n')
alphaset = set('abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ')

def isPropName(name):
    return propre.fullmatch(name) is not None

def isCmdName(name):
    return scmdre.fullmatch(name) is not None

def isUnivName(name):
    return univre.fullmatch(name) is not None

def isFormName(name):
    return formre.fullmatch(name) is not None

def isBasePropNoPivprop(name):
    return basepropnopivpropre.fullmatch(name) is not None

floatre = regex.compile(r'\s*-?\d+(\.\d+)?')

def parse_float(text, off):
    match = floatre.match(text[off:])
    if match is None:
        raise s_exc.BadSyntax(at=off, mesg='Invalid float')

    s = match.group(0)

    return float(s), len(s) + off

def nom(txt, off, cset, trim=True):
    '''
    Consume chars in set from the string and return (subtxt,offset).

    Example:

        text = "foo(bar)"
        chars = set('abcdefghijklmnopqrstuvwxyz')

        name,off = nom(text,0,chars)

    Note:

    This really shouldn't be used for new code
    '''
    if trim:
        while len(txt) > off and txt[off] in whites:
            off += 1

    r = ''
    while len(txt) > off and txt[off] in cset:
        r += txt[off]
        off += 1

    if trim:
        while len(txt) > off and txt[off] in whites:
            off += 1

    return r, off

def meh(txt, off, cset):
    r = ''
    while len(txt) > off and txt[off] not in cset:
        r += txt[off]
        off += 1
    return r, off



import asyncio

import aioimaplib

import synapse.exc as s_exc
import synapse.lib.stormtypes as s_stormtypes

async def run_imap_coro(coro):
    '''
    Raises or returns data
    '''
    try:
        status, data = await coro
    except asyncio.TimeoutError:
        raise s_exc.StormRuntimeError(mesg='Timed out waiting for IMAP server response.') from None

    if status == 'OK':
        return data

    try:
        mesg = data[0].decode()
    except (TypeError, AttributeError, IndexError, UnicodeDecodeError):
        mesg = 'IMAP server returned an error'

    raise s_exc.StormRuntimeError(mesg=mesg, status=status)

@s_stormtypes.registry.registerLib
class ImapLib(s_stormtypes.Lib):
    '''
    A Storm library to connect to an IMAP server.
    '''
    _storm_locals = (
        {
            'name': 'connect',
            'desc': '''
            Open a connection to an IMAP server.

            This method will wait for a "hello" response from the server
            before returning the ``storm:imap:server`` instance.
            ''',
            'type': {
                'type': 'function', '_funcname': 'connect',
                'args': (
                    {'type': 'str', 'name': 'host',
                     'desc': 'The IMAP hostname.'},
                    {'type': 'integer', 'name': 'port', 'default': 993,
                     'desc': 'The IMAP server port.'},
                    {'type': 'int', 'name': 'timeout', 'default': 30,
                     'desc': 'The time to wait for all commands on the server to execute.'},
                    {'type': 'bool', 'name': 'ssl', 'default': True,
                     'desc': 'Use SSL to connect to the IMAP server.'},
                ),
                'returns': {
                    'type': 'storm:imap:server',
                    'desc': 'A new ``storm:imap:server`` instance.'
                },
            },
        },
    )
    _storm_lib_path = ('inet', 'imap', )

    def getObjLocals(self):
        return {
            'connect': self.connect,
        }

    async def connect(self, host, port=993, timeout=30, ssl=True):

        self.runt.confirm(('storm', 'inet', 'imap', 'connect'))

        ssl = await s_stormtypes.tobool(ssl)
        host = await s_stormtypes.tostr(host)
        port = await s_stormtypes.toint(port)
        timeout = await s_stormtypes.toint(timeout, noneok=True)

        if ssl:
            imap_cli = aioimaplib.IMAP4_SSL(host=host, port=port, timeout=timeout)
        else:
            imap_cli = aioimaplib.IMAP4(host=host, port=port, timeout=timeout)

        async def fini():
            # call protocol.logout() so fini() doesn't hang
            await asyncio.wait_for(imap_cli.protocol.logout(), 5)

        self.runt.snap.onfini(fini)

        try:
            await imap_cli.wait_hello_from_server()
        except asyncio.TimeoutError:
            raise s_exc.StormRuntimeError(mesg='Timed out waiting for IMAP server hello.') from None

        return ImapServer(self.runt, imap_cli)

@s_stormtypes.registry.registerType
class ImapServer(s_stormtypes.StormType):
    '''
    An IMAP server for retrieving email messages.
    '''
    _storm_locals = (
        {
            'name': 'list',
            'desc': '''
            List mailbox names.

            By default this method uses a reference_name and pattern to return
            all mailboxes from the root.
            ''',
            'type': {
                'type': 'function', '_funcname': 'list',
                'args': (
                    {'type': 'str', 'name': 'reference_name', 'default': '""',
                     'desc': 'The mailbox reference name.'},
                    {'type': 'str', 'name': 'pattern', 'default': '*',
                     'desc': 'The pattern to filter by.'},
                ),
                'returns': {
                    'type': 'list',
                    'desc': 'An ($ok, $valu) tuple where $valu is a list of names if $ok=True.'
                },
            },
        },
        {
            'name': 'fetch',
            'desc': '''
            Fetch a message by UID in RFC822 format.

            The message is saved to the Axon, and a ``file:bytes`` node is returned.

            Examples:
                Fetch a message, save to the Axon, and yield ``file:bytes`` node::

                    yield $server.fetch("8182")
            ''',
            'type': {
                'type': 'function', '_funcname': 'fetch',
                'args': (
                    {'type': 'str', 'name': 'uid',
                     'desc': 'The single message UID.'},
                ),
                'returns': {
                    'type': 'storm:node',
                    'desc': 'The file:bytes node representing the message.'
                },
            },
        },
        {
            'name': 'login',
            'desc': 'Login to the IMAP server.',
            'type': {
                'type': 'function', '_funcname': 'login',
                'args': (
                    {'type': 'str', 'name': 'user',
                     'desc': 'The username to login with.'},
                    {'type': 'str', 'name': 'passwd',
                     'desc': 'The password to login with.'},
                ),
                'returns': {
                    'type': 'list',
                    'desc': 'An ($ok, $valu) tuple.'
                },
            },
        },
        {
            'name': 'search',
            'desc': '''
            Search for messages using RFC2060 syntax.

            Examples:
                Retrieve all messages::

                    ($ok, $uids) = $server.search("ALL")

                Search by FROM and SINCE::

                    ($ok, $uids) = $server.search("FROM", "visi@vertex.link", "SINCE", "01-Oct-2021")

                Search by a subject substring::

                    ($ok, $uids) = $search.search("HEADER", "Subject", "An email subject")
            ''',
            'type': {
                'type': 'function', '_funcname': 'search',
                'args': (
                    {'type': 'str', 'name': '*args',
                     'desc': 'A set of search criteria to use.'},
                ),
                'returns': {
                    'type': 'list',
                    'desc': 'An ($ok, $valu) tuple, where $valu is a list of UIDs if $ok=True.'
                },
            },
        },
        {
            'name': 'select',
            'desc': 'Select a mailbox to use in subsequent commands.',
            'type': {
                'type': 'function', '_funcname': 'select',
                'args': (
                    {'type': 'str', 'name': 'mailbox', 'default': 'INBOX',
                     'desc': 'The mailbox name to select.'},
                ),
                'returns': {
                    'type': 'list',
                    'desc': 'An ($ok, $valu) tuple.'
                },
            },
        },
        {
            'name': 'markSeen',
            'desc': '''
            Mark messages as seen by an RFC2060 UID message set.

            The command uses the +FLAGS.SILENT command and applies the \\Seen flag.

            Examples:
                Mark a single messsage as seen::

                    ($ok, $valu) = $server.markSeen("8182")

                Mark ranges of messages as seen::

                    ($ok, $valu) = $server.markSeen("1:3,6:9")
            ''',
            'type': {
                'type': 'function', '_funcname': 'markSeen',
                'args': (
                    {'type': 'str', 'name': 'uid_set',
                     'desc': 'The UID message set to apply the flag to.'},
                ),
                'returns': {
                    'type': 'list',
                    'desc': 'An ($ok, $valu) tuple.'
                },
            },
        },
        {
            'name': 'delete',
            'desc': '''
            Mark an RFC2060 UID message as deleted and expunge the mailbox.

            The command uses the +FLAGS.SILENT command and applies the \\Deleted flag.
            The actual behavior of these commands are mailbox configuration dependent.

            Examples:
                Mark a single message as deleted and expunge::

                    ($ok, $valu) = $server.delete("8182")

                Mark ranges of messages as deleted and expunge::

                    ($ok, $valu) = $server.delete("1:3,6:9")
            ''',
            'type': {
                'type': 'function', '_funcname': 'delete',
                'args': (
                    {'type': 'str', 'name': 'uid_set',
                     'desc': 'The UID message set to apply the flag to.'},
                ),
                'returns': {
                    'type': 'list',
                    'desc': 'An ($ok, $valu) tuple.'
                },
            },
        },
    )
    _storm_typename = 'storm:imap:server'

    def __init__(self, runt, imap_cli, path=None):
        s_stormtypes.StormType.__init__(self, path=path)
        self.runt = runt
        self.imap_cli = imap_cli
        self.locls.update(self.getObjLocals())

    def getObjLocals(self):
        return {
            'list': self.list,
            'fetch': self.fetch,
            'login': self.login,
            'delete': self.delete,
            'search': self.search,
            'select': self.select,
            'markSeen': self.markSeen,
        }

    async def login(self, user, passwd):
        user = await s_stormtypes.tostr(user)
        passwd = await s_stormtypes.tostr(passwd)

        coro = self.imap_cli.login(user, passwd)
        await run_imap_coro(coro)

        return True, None

    async def list(self, reference_name='""', pattern='*'):
        pattern = await s_stormtypes.tostr(pattern)
        reference_name = await s_stormtypes.tostr(reference_name)

        coro = self.imap_cli.list(reference_name, pattern)
        data = await run_imap_coro(coro)

        names = []
        for item in data:
            if item == b'Success':
                break
            names.append(item.split(b' ')[-1].decode().strip('"'))

        return True, names

    async def select(self, mailbox='INBOX'):
        mailbox = await s_stormtypes.tostr(mailbox)

        coro = self.imap_cli.select(mailbox=mailbox)
        await run_imap_coro(coro)

        return True, None

    async def search(self, *args):
        args = [await s_stormtypes.tostr(arg) for arg in args]

        coro = self.imap_cli.uid_search(*args)
        data = await run_imap_coro(coro)

        uids = data[0].decode().split(' ') if data[0] else []
        return True, uids

    async def fetch(self, uid):
        '''
        IMAP fetch accepts a message set (e.g. "1", "1:*", "1,2,3"),
        however this method forces fetching a single uid
        to prevent retrieving a very large blob of data.
        '''
        uid = await s_stormtypes.toint(uid)

        await self.runt.snap.core.getAxon()
        axon = self.runt.snap.core.axon

        coro = self.imap_cli.uid('FETCH', str(uid), '(RFC822)')
        data = await run_imap_coro(coro)

        size, sha256b = await axon.put(data[1])

        props = await axon.hashset(sha256b)
        props['size'] = size
        props['mime'] = 'message/rfc822'

        filenode = await self.runt.snap.addNode('file:bytes', props['sha256'], props=props)
        return filenode

    async def delete(self, uid_set):
        uid_set = await s_stormtypes.tostr(uid_set)

        coro = self.imap_cli.uid('STORE', uid_set, '+FLAGS.SILENT (\\Deleted)')
        await run_imap_coro(coro)

        coro = self.imap_cli.expunge()
        await run_imap_coro(coro)

        return True, None

    async def markSeen(self, uid_set):
        uid_set = await s_stormtypes.tostr(uid_set)

        coro = self.imap_cli.uid('STORE', uid_set, '+FLAGS.SILENT (\\Seen)')
        await run_imap_coro(coro)

        return True, None



import synapse.lib.module as s_module

class LangModule(s_module.CoreModule):

    def getModelDefs(self):
        name = 'lang'

        ctors = ()

        forms = (
            ('lang:idiom', {}, (
                ('url', ('inet:url', {}), {
                    'doc': 'Authoritative URL for the idiom.'
                }),
                ('desc:en', ('str', {}), {
                    'doc': 'English description.',
                    'disp': {'hint': 'text'},
                }),
            )),

            ('lang:trans', {}, (
                ('text:en', ('str', {}), {
                    'doc': 'English translation.',
                    'disp': {'hint': 'text'},
                }),
                ('desc:en', ('str', {}), {
                    'doc': 'English description.',
                    'disp': {'hint': 'text'},
                }),
            )),

            ('lang:translation', {}, (
                ('input', ('str', {}), {
                    'ex': 'hola',
                    'doc': 'The input text.',
                }),
                ('input:lang', ('lang:code', {}), {
                    'doc': 'The input language code.'
                }),
                ('output', ('str', {}), {
                    'ex': 'hi',
                    'doc': 'The output text.',
                }),
                ('output:lang', ('lang:code', {}), {
                    'doc': 'The output language code.'
                }),
                ('desc', ('str', {}), {
                    'doc': 'A description of the meaning of the output.',
                    'ex': 'A standard greeting',
                }),
                ('engine', ('it:prod:softver', {}), {
                    'doc': 'The translation engine version used.',
                }),
            ))
        )

        types = (
            ('lang:idiom', ('str', {}), {
                'deprecated': True,
                'doc': 'Deprecated. Please use lang:translation.'
            }),

            ('lang:trans', ('str', {}), {
                'deprecated': True,
                'doc': 'Deprecated. Please use lang:translation.'
            }),

            ('lang:code', ('str', {'lower': True, 'regex': '^[a-z]{2}(.[a-z]{2})?$'}), {
                'ex': 'pt.br',
                'doc': 'An optionally 2 part language code.',
            }),
            ('lang:translation', ('guid', {}), {
                'doc': 'A translation of text from one language to another.',
            }),
        )

        modldef = (name, {
            'ctors': ctors,
            'forms': forms,
            'types': types,
        })
        return (modldef, )



# Python Substrate Interface Library
#
# Copyright 2018-2020 Stichting Polkascan (Polkascan Foundation).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import re


def version_tuple(version_string: str) -> tuple:
    """
    Converts a basic version string to a tuple that can be compared

    Parameters
    ----------
    version_string

    Returns
    -------
    tuple
    """
    if re.search(r'[^\.0-9]', version_string):
        raise ValueError('version_string can only contain numeric characters')

    return tuple(int(v) for v in version_string.split('.'))



import os
import sys
import time
import asyncio
import logging
import argparse

import synapse.exc as s_exc
import synapse.common as s_common
import synapse.cortex as s_cortex
import synapse.telepath as s_telepath

import synapse.lib.cmdr as s_cmdr
import synapse.lib.output as s_output
import synapse.lib.msgpack as s_msgpack
import synapse.lib.version as s_version
import synapse.lib.encoding as s_encoding

logger = logging.getLogger(__name__)

reqver = '>=0.2.0,<3.0.0'

def getItems(*paths):
    items = []
    for path in paths:
        if path.endswith('.json'):
            item = s_common.jsload(path)
            if not isinstance(item, list):
                item = [item]
            items.append((path, item))
        elif path.endswith('.jsonl'):
            with s_common.genfile(path) as fd:
                item = list(s_encoding.iterdata(fd, False, format='jsonl'))
                items.append((path, item))
        elif path.endswith(('.yaml', '.yml')):
            item = s_common.yamlload(path)
            if not isinstance(item, list):
                item = [item]
            items.append((path, item))
        elif path.endswith('.mpk') or path.endswith('.nodes'):
            genr = s_msgpack.iterfile(path)
            items.append((path, genr))
        else:  # pragma: no cover
            logger.warning('Unsupported file path: [%s]', path)
    return items

async def addFeedData(core, outp, feedformat, debug=False, *paths, chunksize=1000, offset=0, viewiden=None):

    items = getItems(*paths)
    for path, item in items:

        bname = os.path.basename(path)

        tick = time.time()
        outp.printf(f'Adding items from [{path}]')

        foff = 0
        for chunk in s_common.chunks(item, chunksize):

            clen = len(chunk)
            if offset and foff + clen < offset:
                # We have not yet encountered a chunk which
                # will include the offset size.
                foff += clen
                continue

            await core.addFeedData(feedformat, chunk, viewiden=viewiden)

            foff += clen
            outp.printf(f'Added [{clen}] items from [{bname}] - offset [{foff}]')

        tock = time.time()

        outp.printf(f'Done consuming from [{bname}]')
        outp.printf(f'Took [{tock - tick}] seconds.')

    if debug:
        await s_cmdr.runItemCmdr(core, outp, True)

async def main(argv, outp=None):

    if outp is None:  # pragma: no cover
        outp = s_output.OutPut()

    pars = makeargparser()
    opts = pars.parse_args(argv)

    if opts.offset:
        if len(opts.files) > 1:
            outp.printf('Cannot start from a arbitrary offset for more than 1 file.')
            return 1

        outp.printf(f'Starting from offset [{opts.offset}] - it may take a while'
                    f' to get to that location in the input file.')

    if opts.test:
        async with s_cortex.getTempCortex(mods=opts.modules) as prox:
            await addFeedData(prox, outp, opts.format, opts.debug,
                        chunksize=opts.chunksize,
                        offset=opts.offset,
                        *opts.files)

    elif opts.cortex:
        async with await s_telepath.openurl(opts.cortex) as core:
            try:
                s_version.reqVersion(core._getSynVers(), reqver)
            except s_exc.BadVersion as e:
                valu = s_version.fmtVersion(*e.get('valu'))
                outp.printf(f'Cortex version {valu} is outside of the feed tool supported range ({reqver}).')
                outp.printf(f'Please use a version of Synapse which supports {valu}; '
                      f'current version is {s_version.verstring}.')
                return 1
            await addFeedData(core, outp, opts.format, opts.debug,
                              chunksize=opts.chunksize,
                              offset=opts.offset, viewiden=opts.view,
                              *opts.files)

    else:  # pragma: no cover
        outp.printf('No valid options provided [%s]', opts)
        return 1

    return 0

def makeargparser():
    desc = 'Command line tool for ingesting data into a cortex'
    pars = argparse.ArgumentParser('synapse.tools.feed', description=desc)

    muxp = pars.add_mutually_exclusive_group(required=True)
    muxp.add_argument('--cortex', '-c', type=str,
                      help='Cortex to connect and add nodes too.')
    muxp.add_argument('--test', '-t', default=False, action='store_true',
                      help='Perform a local ingest against a temporary cortex.')

    pars.add_argument('--debug', '-d', default=False, action='store_true',
                      help='Drop to interactive prompt to inspect cortex after loading data.')
    pars.add_argument('--format', '-f', type=str, action='store', default='syn.nodes',
                      help='Feed format to use for the ingested data.')
    pars.add_argument('--modules', '-m', type=str, action='append', default=[],
                      help='Additional modules to load locally with a test Cortex.')
    pars.add_argument('--chunksize', type=int, action='store', default=1000,
                      help='Default chunksize for iterating over items.')
    pars.add_argument('--offset', type=int, action='store', default=0,
                      help='Item offset to start consuming data from.')
    pars.add_argument('--view', type=str, action='store', default=None,
                      help='The View to ingest the data into.')
    pars.add_argument('files', nargs='*', help='json/yaml/msgpack feed files')

    return pars

if __name__ == '__main__':  # pragma: no cover
    s_common.setlogging(logger, 'DEBUG')
    asyncio.run(main(sys.argv[1:]))



import types

import synapse.common as s_common

def hashitem(item):
    '''
    Generate a uniq hash for the JSON compatible primitive data structure.
    '''
    norm = normitem(item)
    return s_common.guid(norm)

def normitem(item):
    normer = normers.get(type(item))
    if normer:
        return normer(item)

    return item

def normdict(item):
    return tuple(sorted([(normitem(key), normitem(val)) for key, val in item.items() if val is not None]))

def normiter(item):
    return tuple([normitem(i) for i in item if i is not None])

normers = {
    dict: normdict,
    list: normiter,
    tuple: normiter,
    types.GeneratorType: normiter,
}



import synapse.lib.module as s_module

class RiskModule(s_module.CoreModule):

    def getModelDefs(self):

        modl = {
            'types': (
                ('risk:vuln', ('guid', {}), {
                    'doc': 'A unique vulnerability.',
                }),
                ('risk:hasvuln', ('guid', {}), {
                    'doc': 'An instance of a vulnerability present in a target.',
                }),
                ('risk:attack', ('guid', {}), {
                    'doc': 'An instance of an actor attacking a target.'
                }),
                ('risk:alert', ('guid', {}), {
                    'doc': 'An instance of an alert which indicates the presence of a risk.',
                }),
                ('risk:compromise', ('guid', {}), {
                    'doc': 'An instance of a compromise and its aggregate impact.',
                }),
                ('risk:mitigation', ('guid', {}), {
                    'doc': 'A mitigation for a specific risk:vuln.',
                }),
                ('risk:attacktype', ('taxonomy', {}), {
                    'doc': 'An attack type taxonomy.',
                    'interfaces': ('taxonomy',),
                }),
                ('risk:compromisetype', ('taxonomy', {}), {
                    'doc': 'A compromise type taxonomy.',
                    'ex': 'cno.breach',
                    'interfaces': ('taxonomy',),
                }),
            ),
            'forms': (
                ('risk:mitigation', {}, (
                    ('vuln', ('risk:vuln', {}), {
                        'doc': 'The vulnerability that this mitigation addresses.'}),
                    ('name', ('str', {}), {
                        'doc': 'A brief name for this risk mitigation.'}),
                    ('desc', ('str', {}), {
                        'disp': {'hint': 'text'},
                        'doc': 'A description of the mitigation approach for the vulnerability.'}),
                    ('software', ('it:prod:softver', {}), {
                        'doc': 'A software version which implements a fix for the vulnerability.'}),
                    ('hardware', ('it:prod:hardware', {}), {
                        'doc': 'A hardware version which implements a fix for the vulnerability.'}),
                )),
                ('risk:vuln', {}, (
                    ('name', ('str', {}), {
                        'doc': 'A user specified name for the vulnerability.',
                    }),
                    ('type', ('str', {}), {
                        'doc': 'A user specified type for the vulnerability.',
                    }),
                    ('desc', ('str', {}), {
                        'doc': 'A description of the vulnerability.',
                        'disp': {'hint': 'text'},
                    }),
                    ('cve', ('it:sec:cve', {}), {
                        'doc': 'The CVE ID of the vulnerability.',
                    }),
                    ('cvss:av', ('str', {'enums': 'N,A,V,L'}), {
                        'doc': 'The CVSS Attack Vector (AV) value.',
                    }),
                    ('cvss:ac', ('str', {'enums': 'L,H'}), {
                        'doc': 'The CVSS Attack Complexity (AC) value.',
                        'disp': {'enums': (('Low', 'L'), ('High', 'H'))},
                    }),
                    ('cvss:pr', ('str', {'enums': 'N,L,H'}), {
                        'doc': 'The CVSS Privileges Required (PR) value.',
                        'disp': {'enums': (
                            {'title': 'None', 'value': 'N', 'doc': 'FIXME privs stuff'},
                            {'title': 'Low', 'value': 'L', 'doc': 'FIXME privs stuff'},
                            {'title': 'High', 'value': 'H', 'doc': 'FIXME privs stuff'},
                        )},
                    }),
                    ('cvss:ui', ('str', {'enums': 'N,R'}), {
                        'doc': 'The CVSS User Interaction (UI) value.',
                    }),
                    ('cvss:s', ('str', {'enums': 'U,C'}), {
                        'doc': 'The CVSS Scope (S) value.',
                    }),
                    ('cvss:c', ('str', {'enums': 'N,L,H'}), {
                        'doc': 'The CVSS Confidentiality Impact (C) value.',
                    }),
                    ('cvss:i', ('str', {'enums': 'N,L,H'}), {
                        'doc': 'The CVSS Integrity Impact (I) value.',
                    }),
                    ('cvss:a', ('str', {'enums': 'N,L,H'}), {
                        'doc': 'The CVSS Availability Impact (A) value.',
                    }),
                    ('cvss:e', ('str', {'enums': 'X,U,P,F,H'}), {
                        'doc': 'The CVSS Exploit Code Maturity (E) value.',
                    }),
                    ('cvss:rl', ('str', {'enums': 'X,O,T,W,U'}), {
                        'doc': 'The CVSS Remediation Level (RL) value.',
                    }),
                    ('cvss:rc', ('str', {'enums': 'X,U,R,C'}), {
                        'doc': 'The CVSS Report Confidence (AV) value.',
                    }),
                    ('cvss:mav', ('str', {'enums': 'X,N,A,L,P'}), {
                        'doc': 'The CVSS Environmental Attack Vector (MAV) value.',
                    }),
                    ('cvss:mac', ('str', {'enums': 'X,L,H'}), {
                        'doc': 'The CVSS Environmental Attack Complexity (MAC) value.',
                    }),
                    ('cvss:mpr', ('str', {'enums': 'X,N,L,H'}), {
                        'doc': 'The CVSS Environmental Privileges Required (MPR) value.',
                    }),
                    ('cvss:mui', ('str', {'enums': 'X,N,R'}), {
                        'doc': 'The CVSS Environmental User Interaction (MUI) value.',
                    }),
                    ('cvss:ms', ('str', {'enums': 'X,U,C'}), {
                        'doc': 'The CVSS Environmental Scope (MS) value.',
                    }),
                    ('cvss:mc', ('str', {'enums': 'X,N,L,H'}), {
                        'doc': 'The CVSS Environmental Confidentiality Impact (MC) value.',
                    }),
                    ('cvss:mi', ('str', {'enums': 'X,N,L,H'}), {
                        'doc': 'The CVSS Environmental Integrity Impact (MI) value.',
                    }),
                    ('cvss:ma', ('str', {'enums': 'X,N,L,H'}), {
                        'doc': 'The CVSS Environmental Accessibility Impact (MA) value.',
                    }),
                    ('cvss:cr', ('str', {'enums': 'X,L,M,H'}), {
                        'doc': 'The CVSS Environmental Confidentiality Requirement (CR) value.',
                    }),
                    ('cvss:ir', ('str', {'enums': 'X,L,M,H'}), {
                        'doc': 'The CVSS Environmental Integrity Requirement (IR) value.',
                    }),
                    ('cvss:ar', ('str', {'enums': 'X,L,M,H'}), {
                        'doc': 'The CVSS Environmental Availability Requirement (AR) value.',
                    }),
                    ('cvss:score', ('float', {}), {
                        'doc': 'The Overall CVSS Score value.',
                    }),
                    ('cvss:score:base', ('float', {}), {
                        'doc': 'The CVSS Base Score value.',
                    }),
                    ('cvss:score:temporal', ('float', {}), {
                        'doc': 'The CVSS Temporal Score value.',
                    }),
                    ('cvss:score:environmental', ('float', {}), {
                        'doc': 'The CVSS Environmental Score value.',
                    }),
                    ('cwes', ('array', {'type': 'it:sec:cwe', 'uniq': True, 'sorted': True}), {
                        'doc': 'An array of MITRE CWE values that apply to the vulnerability.',
                    }),
                )),

                ('risk:hasvuln', {}, (
                    ('vuln', ('risk:vuln', {}), {
                        'doc': 'The vulnerability present in the target.'
                    }),
                    ('person', ('ps:person', {}), {
                        'doc': 'The vulnerable person.',
                    }),
                    ('org', ('ou:org', {}), {
                        'doc': 'The vulnerable org.',
                    }),
                    ('place', ('geo:place', {}), {
                        'doc': 'The vulnerable place.',
                    }),
                    ('software', ('it:prod:softver', {}), {
                        'doc': 'The vulnerable software.',
                    }),
                    ('hardware', ('it:prod:hardware', {}), {
                        'doc': 'The vulnerable hardware.',
                    }),
                    ('spec', ('mat:spec', {}), {
                        'doc': 'The vulnerable material specification.',
                    }),
                    ('item', ('mat:item', {}), {
                        'doc': 'The vulnerable material item.',
                    }),
                    ('host', ('it:host', {}), {
                        'doc': 'The vulnerable host.'
                    })
                )),

                ('risk:alert', {}, (
                    ('type', ('str', {'lower': True, 'onespace': True, 'strip': True}), {
                        'doc': 'An alert type.',
                    }),
                    ('name', ('str', {}), {
                        'doc': 'The alert name.',
                    }),
                    ('desc', ('str', {}), {
                        'disp': {'hint': 'text'},
                        'doc': 'A free-form description / overview of the alert.',
                    }),
                    ('detected', ('time', {}), {
                        'doc': 'The time the alerted condition was detected.',
                    }),
                    ('vuln', ('risk:vuln', {}), {
                        'doc': 'The optional vulnerability that the alert indicates.',
                    }),
                    ('attack', ('risk:attack', {}), {
                        'doc': 'A confirmed attack that this alert indicates.',
                    }),
                )),
                ('risk:compromisetype', {}, ()),
                ('risk:compromise', {}, (
                    ('name', ('str', {'lower': True, 'onespace': True, 'strip': True}), {
                        'doc': 'A brief name for the compromise event.',
                    }),
                    ('desc', ('str', {}), {
                        'disp': {'hint': 'text'},
                        'doc': 'A prose description of the compromise event.',
                    }),
                    ('type', ('risk:compromisetype', {}), {
                        'ex': 'cno.breach',
                        'doc': 'The compromise type.',
                    }),
                    ('target', ('ps:contact', {}), {
                        'doc': 'Contact information of the target.',
                    }),
                    ('attacker', ('ps:contact', {}), {
                        'doc': 'Contact information of the attacker.',
                    }),
                    ('campaign', ('ou:campaign', {}), {
                        'doc': 'The campaign that this compromise is part of.',
                    }),
                    ('time', ('time', {}), {
                        'doc': 'Earliest known evidence of compromise.',
                    }),
                    ('lasttime', ('time', {}), {
                        'doc': 'Last known evidence of compromise.',
                    }),
                    ('duration', ('duration', {}), {
                        'doc': 'The duration of the compromise.',
                    }),
                    ('loss:pii', ('int', {}), {
                        'doc': 'The number of records compromised which contain PII.',
                    }),
                    ('loss:econ', ('econ:price', {}), {
                        'doc': 'The total economic cost of the compromise.',
                    }),
                    ('loss:life', ('int', {}), {
                        'doc': 'The total loss of life due to the compromise.',
                    }),
                    ('loss:bytes', ('int', {}), {
                        'doc': 'An estimate of the volume of data compromised.',
                    }),
                    ('ransom:paid', ('econ:price', {}), {
                        'doc': 'The value of the ransom paid by the target.',
                    }),
                    ('ransom:price', ('econ:price', {}), {
                        'doc': 'The value of the ransom demanded by the attacker.',
                    }),
                    ('response:cost', ('econ:price', {}), {
                        'doc': 'The economic cost of the response and mitigation efforts.',
                    }),
                    ('theft:price', ('econ:price', {}), {
                        'doc': 'The total value of the theft of assets.',
                    }),
                    ('econ:currency', ('econ:currency', {}), {
                        'doc': 'The currency type for the econ:price fields.',
                    }),
                    # -(stole)> file:bytes ps:contact file:bytes
                    # -(compromised)> geo:place it:account it:host
                )),
                ('risk:attacktype', {}, ()),
                ('risk:attack', {}, (
                    ('desc', ('str', {}), {
                        'doc': 'A description of the attack.',
                        'disp': {'hint': 'text'},
                    }),
                    ('type', ('risk:attacktype', {}), {
                        'ex': 'cno.phishing',
                        'doc': 'The attack type.',
                    }),
                    ('time', ('time', {}), {
                        'doc': 'Set if the time of the attack is known.',
                    }),
                    ('success', ('bool', {}), {
                        'doc': 'Set if the attack was known to have succeeded or not.',
                    }),
                    ('targeted', ('bool', {}), {
                        'doc': 'Set if the attack was assessed to be targeted or not.',
                    }),
                    ('goal', ('ou:goal', {}), {
                        'doc': 'The tactical goal of this specific attack.',
                    }),
                    ('campaign', ('ou:campaign', {}), {
                        'doc': 'Set if the attack was part of a larger campaign.',
                    }),
                    ('compromise', ('risk:compromise', {}), {
                        'doc': 'A compromise that this attack contributed to.',
                    }),
                    ('prev', ('risk:attack', {}), {
                        'doc': 'The previous/parent attack in a list or hierarchy.',
                    }),
                    ('actor:org', ('ou:org', {}), {
                        'deprecated': True,
                        'doc': 'Deprecated. Please use :attacker to allow entity resolution.',
                    }),
                    ('actor:person', ('ps:person', {}), {
                        'deprecated': True,
                        'doc': 'Deprecated. Please use :attacker to allow entity resolution.',
                    }),
                    ('attacker', ('ps:contact', {}), {
                        'doc': 'Contact information associated with the attacker.',
                    }),
                    ('target', ('ps:contact', {}), {
                        'doc': 'Contact information associated with the target.',
                    }),
                    ('target:org', ('ou:org', {}), {
                        'deprecated': True,
                        'doc': 'Deprecated. Please use :target to allow entity resolution.',
                    }),
                    ('target:host', ('it:host', {}), {
                        'doc': 'The host was the target of the attack.',
                    }),
                    ('target:person', ('ps:person', {}), {
                        'deprecated': True,
                        'doc': 'Deprecated. Please use :target to allow entity resolution.',
                    }),
                    ('target:place', ('geo:place', {}), {
                        'doc': 'The place that was the target of the attack.',
                    }),
                    ('via:ipv4', ('inet:ipv4', {}), {
                        'doc': 'The target host was contacted via the IPv4 address.',
                    }),
                    ('via:ipv6', ('inet:ipv6', {}), {
                        'doc': 'The target host was contacted via the IPv6 address.',
                    }),
                    ('via:email', ('inet:email', {}), {
                        'doc': 'The target person/org was contacted via the email address.',
                    }),
                    ('via:phone', ('tel:phone', {}), {
                        'doc': 'The target person/org was contacted via the phone number.',
                    }),
                    ('used:vuln', ('risk:vuln', {}), {
                        'doc': 'The actor used the vuln in the attack.',
                    }),
                    ('used:url', ('inet:url', {}), {
                        'doc': 'The actor used the url in the attack.',
                    }),
                    ('used:host', ('it:host', {}), {
                        'doc': 'The actor used the host in the attack.',
                    }),
                    ('used:email', ('inet:email', {}), {
                        'doc': 'The actor used the email in the attack.',
                    }),
                    ('used:file', ('file:bytes', {}), {
                        'doc': 'The actor used the file in the attack.',
                    }),
                    ('used:server', ('inet:server', {}), {
                        'doc': 'The actor used the server in the attack.',
                    }),
                    ('used:software', ('it:prod:softver', {}), {
                        'doc': 'The actor used the software in the attack.',
                    }),
                )),
            ),
        }
        name = 'risk'
        return ((name, modl), )



import sys
import logging
import argparse

import synapse.telepath as s_telepath

import synapse.lib.output as s_output

logger = logging.getLogger(__name__)

def main(argv, outp=s_output.stdout):

    pars = argparse.ArgumentParser(prog='cryo.list', description='List tanks within a cryo cell.')
    pars.add_argument('cryocell', nargs='+', help='Telepath URLs to cryo cells.')

    opts = pars.parse_args(argv)

    for url in opts.cryocell:

        outp.printf(url)

        with s_telepath.openurl(url) as cryo:

            for name, info in cryo.list():

                outp.printf(f'    {name}: {info}')

    return 0

if __name__ == '__main__':  # pragma: no cover
    logging.basicConfig()
    sys.exit(main(sys.argv[1:]))



import synapse.exc as s_exc
import synapse.tests.utils as s_test

yaml00 = '''
foo: bar
baz: faz
'''

class YamlTest(s_test.SynTest):

    async def test_stormlib_yaml(self):
        async with self.getTestCore() as core:
            valu = await core.callStorm('''
                return($lib.yaml.save((hehe, haha, hoho)))
            ''')
            self.eq(valu, '- hehe\n- haha\n- hoho\n')

            valu = await core.callStorm('''
                return($lib.yaml.load($yaml00))
            ''', opts={'vars': {'yaml00': yaml00}})
            self.eq(valu, {'foo': 'bar', 'baz': 'faz'})

            valu = await core.callStorm('''
                return($lib.yaml.save($lib.yaml.load($yaml00)))
            ''', opts={'vars': {'yaml00': yaml00}})
            self.eq(valu, 'baz: faz\nfoo: bar\n')

            valu = await core.callStorm('''
                return($lib.yaml.save($lib.yaml.load($yaml00), sort_keys=$lib.false))
            ''', opts={'vars': {'yaml00': yaml00}})
            self.eq(valu, 'foo: bar\nbaz: faz\n')

            with self.raises(s_exc.BadArg):
                opts = {'vars': {'bytes': b'foo: bar: baz:'}}
                await core.callStorm('$lib.yaml.load($bytes)', opts=opts)

            with self.raises(s_exc.NoSuchType):
                await core.callStorm('$lib.yaml.save($lib.queue.gen(foo))', opts=opts)



import os

import synapse.common as s_common
import synapse.tests.utils as s_t_utils

class VendorTest(s_t_utils.SynTest):

    def setUp(self) -> None:
        if not s_common.envbool('SYN_VENDOR_TEST'):  # pragma: no cover
            self.skip('Skipping vendored test.')



import asyncio

import synapse.lib.base as s_base
import synapse.lib.task as s_task

class Boss(s_base.Base):

    '''
    An object to track "promoted" async tasks.
    '''
    async def __anit__(self):
        await s_base.Base.__anit__(self)
        self.tasks = {}
        self.onfini(self._onBossFini)

    async def _onBossFini(self):
        for task in list(self.tasks.values()):
            await task.kill()

    def ps(self):
        # top level tasks only...
        return [t for t in self.tasks.values() if t.root is None]

    def get(self, iden):
        return self.tasks.get(iden)

    async def promote(self, name, user, info=None):
        '''
        Promote the currently running task.

        Args:
            name (str): The name of the task.
            user: The User who owns the task.
            info: An optional information dictionary containing information about the task.

        Returns:
            s_task.Task: The Synapse Task object.
        '''
        task = asyncio.current_task()

        synt = getattr(task, '_syn_task', None)

        if synt is not None:

            if synt.root is None:
                return synt

            synt.root.kids.pop(synt.iden)
            synt.root = None
            return synt

        return await s_task.Task.anit(self, task, name, user, info=info)

    async def execute(self, coro, name, user, info=None):
        '''
        Create a synapse task from the given coroutine.
        '''
        task = self.schedCoro(coro)
        return await s_task.Task.anit(self, task, name, user, info=info)



import os
import copy
import json
import pprint
import logging
import contextlib
import collections

import vcr
import regex

from unittest import mock

import synapse.exc as s_exc
import synapse.common as s_common

import synapse.lib.base as s_base
import synapse.lib.output as s_output
import synapse.lib.dyndeps as s_dyndeps
import synapse.lib.stormhttp as s_stormhttp

import synapse.cmds.cortex as s_cmds_cortex

import synapse.tools.storm as s_storm
import synapse.tools.genpkg as s_genpkg


re_directive = regex.compile(r'^\.\.\s(storm.*|[^:])::(?:\s(.*)$|$)')

logger = logging.getLogger(__name__)


class OutPutRst(s_output.OutPutStr):
    '''
    Rst specific helper for output intended to be indented
    in RST text as a literal block.
    '''
    prefix = '    '

    def printf(self, mesg, addnl=True):

        if '\n' in mesg:
            logger.debug(f'Newline found in [{mesg}]')
            parts = mesg.split('\n')
            mesg0 = '\n'.join([self.prefix + part for part in parts[1:]])
            mesg = '\n'.join((parts[0], mesg0))

        return s_output.OutPutStr.printf(self, mesg, addnl)


class StormOutput(s_cmds_cortex.StormCmd):
    '''
    Produce standard output from a stream of storm runtime messages.
    Must be instantiated for a single query with a rstorm context.
    '''

    _cmd_syntax = (
        ('--hide-query', {}),
    ) + s_cmds_cortex.StormCmd._cmd_syntax

    def __init__(self, core, ctx, stormopts=None, opts=None):
        if opts is None:
            opts = {}

        s_cmds_cortex.StormCmd.__init__(self, None, **opts)

        self.stormopts = stormopts or {}

        # hide a few mesg types by default
        for mtype in ('init', 'fini', 'node:edits', 'node:edits:count', 'prov:new'):
            self.cmdmeths[mtype] = self._silence

        self.core = core
        self.ctx = ctx
        self.lines = []
        self.prefix = '    '

    async def runCmdLine(self, line):
        opts = self.getCmdOpts(f'storm {line}')
        return await self.runCmdOpts(opts)

    def _printNodeProp(self, name, valu):
        base = f'        {name} = '
        if '\n' in valu:
            parts = collections.deque(valu.split('\n'))
            ws = ' ' * len(base)
            self.printf(f'{base}{parts.popleft()}')
            while parts:
                part = parts.popleft()
                self.printf(f'{ws}{part}')

        else:
            self.printf(f'{base}{valu}')

    async def _mockHttp(self, *args, **kwargs):
        info = {
            'code': 200,
            'body': '{}',
        }

        resp = self.ctx.get('mock-http')
        if resp:
            body = resp.get('body')

            if isinstance(body, (dict, list)):
                body = json.dumps(body)

            info = {
                'code': resp.get('code', 200),
                'body': body.encode(),
            }

        return s_stormhttp.HttpResp(info)

    @contextlib.contextmanager
    def _shimHttpCalls(self, vcr_kwargs):
        path = self.ctx.get('mock-http-path')
        if not vcr_kwargs:
            vcr_kwargs = {}

        if path:
            path = os.path.abspath(path)
            # try it as json first (since yaml can load json...). if it parses, we're old school
            # if it doesn't, either it doesn't exist/we can't read it/we can't parse it.
            # in any of those cases, default to using vcr
            try:
                with open(path, 'r') as fd:
                    byts = json.load(fd)
            except (FileNotFoundError, json.decoder.JSONDecodeError):
                byts = None

            if not byts:
                recorder = vcr.VCR(**vcr_kwargs)
                vcrcb = self.ctx.get('storm-vcr-callback', None)
                if vcrcb:
                    vcrcb(recorder)
                with recorder.use_cassette(os.path.abspath(path)) as cass:
                    yield cass
                    self.ctx.pop('mock-http-path', None)
            else:  # backwards compat
                if not os.path.isfile(path):
                    raise s_exc.NoSuchFile(mesg='Storm HTTP mock filepath does not exist', path=path)
                self.ctx['mock-http'] = byts
                with mock.patch('synapse.lib.stormhttp.LibHttp._httpRequest', new=self._mockHttp):
                    yield
        else:
            yield

    def printf(self, mesg, addnl=True, color=None):
        line = f'{self.prefix}{mesg}'
        if '\n' in line:
            logger.debug(f'Newline found in [{mesg}]')
            parts = line.split('\n')
            mesg0 = '\n'.join([self.prefix + part for part in parts[1:]])
            line = '\n'.join((parts[0], mesg0))

        self.lines.append(line)
        return line

    def _silence(self, mesg, opts):
        pass

    def _onErr(self, mesg, opts):
        # raise on err for rst
        if self.ctx.pop('storm-fail', None):
            s_cmds_cortex.StormCmd._onErr(self, mesg, opts)
            return
        raise s_exc.StormRuntimeError(mesg=mesg)

    async def runCmdOpts(self, opts):

        text = opts.get('query')

        if not opts.get('hide-query'):
            self.printf(f'> {text}')

        stormopts = copy.deepcopy(self.stormopts)

        stormopts.setdefault('repr', True)
        stormopts.setdefault('path', opts.get('path', False))

        hide_unknown = True

        # Let this raise on any errors
        with self._shimHttpCalls(self.ctx.get('storm-vcr-opts')):
            async for mesg in self.core.storm(text, opts=stormopts):

                if opts.get('debug'):
                    self.printf(pprint.pformat(mesg))
                    continue

                try:
                    func = self.cmdmeths[mesg[0]]
                except KeyError:  # pragma: no cover
                    if hide_unknown:
                        continue
                    self.printf(repr(mesg), color=s_cmds_cortex.UNKNOWN_COLOR)
                else:
                    func(mesg, opts)

        return '\n'.join(self.lines)

class StormCliOutput(s_storm.StormCli):

    async def __anit__(self, item, outp=s_output.stdout, opts=None):
        await s_storm.StormCli.__anit__(self, item, outp, opts)
        self.ctx = {}
        self._print_skips.append('init')
        self._print_skips.append('fini')
        self._print_skips.append('prov:new')
        self._print_skips.append('node:edits')
        self._print_skips.append('node:edits:count')

    def printf(self, mesg, addnl=True, color=None):
        mesg = f'    {mesg}'
        s_storm.StormCli.printf(self, mesg, addnl, color)

    async def handleErr(self, mesg):
        #  raise on err for rst
        if self.ctx.pop('storm-fail', None):
            await s_storm.StormCli.handleErr(self, mesg)
            return
        raise s_exc.StormRuntimeError(mesg=mesg)

    def _printNodeProp(self, name, valu):
        base = f'        {name} = '
        if '\n' in valu:
            parts = collections.deque(valu.split('\n'))
            ws = ' ' * len(base)
            self.printf(f'{base}{parts.popleft()}')
            while parts:
                part = parts.popleft()
                self.printf(f'{ws}{part}')

        else:
            self.printf(f'{base}{valu}')

    async def _mockHttp(self, *args, **kwargs):
        info = {
            'code': 200,
            'body': '{}',
        }

        resp = self.ctx.get('mock-http')
        if resp:
            body = resp.get('body')

            if isinstance(body, (dict, list)):
                body = json.dumps(body)

            info = {
                'code': resp.get('code', 200),
                'body': body.encode(),
            }

        return s_stormhttp.HttpResp(info)

    @contextlib.contextmanager
    def _shimHttpCalls(self, vcr_kwargs):
        path = self.ctx.get('mock-http-path')
        if not vcr_kwargs:
            vcr_kwargs = {}

        if path:
            path = os.path.abspath(path)
            # try it as json first (since yaml can load json...). if it parses, we're old school
            # if it doesn't, either it doesn't exist/we can't read it/we can't parse it.
            # in any of those cases, default to using vcr
            try:
                with open(path, 'r') as fd:
                    byts = json.load(fd)
            except (FileNotFoundError, json.decoder.JSONDecodeError):
                byts = None

            if not byts:
                with vcr.use_cassette(os.path.abspath(path), **vcr_kwargs) as cass:
                    yield cass
                    self.ctx.pop('mock-http-path', None)
            else:  # backwards compat
                if not os.path.isfile(path):
                    raise s_exc.NoSuchFile(mesg='Storm HTTP mock filepath does not exist', path=path)
                self.ctx['mock-http'] = byts
                with mock.patch('synapse.lib.stormhttp.LibHttp._httpRequest', new=self._mockHttp):
                    yield
        else:
            yield

    async def runRstCmdLine(self, text, ctx, stormopts=None):
        self.ctx = ctx

        self.printf(self.cmdprompt + text)

        with self._shimHttpCalls(self.ctx.get('storm-vcr-opts')):

            await self.runCmdLine(text, opts=stormopts)

        return str(self.outp)

@contextlib.asynccontextmanager
async def getCell(ctor, conf):
    loc = s_dyndeps.getDynLocal(ctor)
    if loc is None:
        raise s_exc.NoSuchCtor(mesg=f'Unable to resolve ctor [{ctor}]', ctor=ctor)
    with s_common.getTempDir() as dirn:
        async with await loc.anit(dirn, conf=conf) as cell:
            yield cell

class StormRst(s_base.Base):

    async def __anit__(self, rstfile):
        await s_base.Base.__anit__(self)

        self.rstfile = s_common.genpath(rstfile)

        if not os.path.isfile(rstfile):
            raise s_exc.BadConfValu(mesg='A valid rstfile must be specified', rstfile=self.rstfile)

        self.linesout = []
        self.context = {}
        self.stormvars = {}

        self.core = None

        self.handlers = {
            'storm': self._handleStorm,
            'storm-cli': self._handleStormCli,
            'storm-pkg': self._handleStormPkg,
            'storm-pre': self._handleStormPre,
            'storm-svc': self._handleStormSvc,
            'storm-fail': self._handleStormFail,
            'storm-opts': self._handleStormOpts,
            'storm-cortex': self._handleStormCortex,
            'storm-envvar': self._handleStormEnvVar,
            'storm-expect': self._handleStormExpect,
            'storm-multiline': self._handleStormMultiline,
            'storm-mock-http': self._handleStormMockHttp,
            'storm-vcr-opts': self._handleStormVcrOpts,
            'storm-clear-http': self._handleStormClearHttp,
            'storm-vcr-callback': self._handleStormVcrCallback,
        }

    async def _getCell(self, ctor, conf=None):
        if conf is None:
            conf = {}

        cell = await self.enter_context(getCell(ctor, conf))

        return cell

    def _printf(self, line):
        self.linesout.append(line)

    def _reqCore(self):
        if self.core is None:
            mesg = 'No cortex set.  Use .. storm-cortex::'
            raise s_exc.NoSuchVar(mesg=mesg)
        return self.core

    def _getHandler(self, directive):
        handler = self.handlers.get(directive)
        if handler is None:
            raise s_exc.NoSuchName(mesg=f'The {directive} directive is not supported', directive=directive)
        return handler

    async def _handleStorm(self, text):
        '''
        Run a Storm command and generate text from the output.

        Args:
            text (str): A valid Storm query.
        '''
        core = self._reqCore()
        text = self._getStormMultiline(text)

        self._printf('::\n')
        self._printf('\n')

        soutp = StormOutput(core, self.context, stormopts=self.context.get('storm-opts'))
        self._printf(await soutp.runCmdLine(text))

        if self.context.pop('storm-fail', None):
            raise s_exc.StormRuntimeError(mesg='Expected a failure, but none occurred.')

        self._printf('\n\n')

    async def _handleStormCli(self, text):
        core = self._reqCore()
        outp = OutPutRst()
        text = self._getStormMultiline(text)

        self._printf('::\n')
        self._printf('\n')

        cli = await StormCliOutput.anit(item=core, outp=outp)

        self._printf(await cli.runRstCmdLine(text, self.context, stormopts=self.context.get('storm-opts')))

        if self.context.pop('storm-fail', None):
            raise s_exc.StormRuntimeError(mesg='Expected a failure, but none occurred.')

        self._printf('\n')

    async def _handleStormPkg(self, text):
        '''
        Load a Storm package into the Cortex by path.

        Args:
            text (str): The path to a Storm package YAML file.
        '''
        if not os.path.isfile(text):
            raise s_exc.NoSuchFile(mesg='Storm Package filepath does not exist', path=text)

        core = self._reqCore()

        pkg = s_genpkg.loadPkgProto(text)
        await core.addStormPkg(pkg)

    async def _handleStormPre(self, text):
        '''
        Run a Storm query to prepare the Cortex without output.

        Args:
            text (str): A valid Storm query
        '''
        core = self._reqCore()

        self.context.setdefault('storm-opts', {})

        stormopts = self.context.get('storm-opts')
        stormopts.setdefault('vars', {})

        # only map env vars in for storm-pre
        stormopts = copy.deepcopy(stormopts)
        stormopts['vars'].update(self.stormvars)

        soutp = StormOutput(core, self.context, stormopts=stormopts)
        await soutp.runCmdLine(text)

        self.context.pop('storm-fail', None)

    async def _handleStormSvc(self, text):
        '''
        Load a Storm service by ctor and add to the Cortex.

        Args:
            text (str): <ctor> <svcname> <optional JSON string to use as svcconf>
        '''
        core = self._reqCore()

        splts = text.split(' ', 2)
        ctor, svcname = splts[:2]
        svcconf = json.loads(splts[2].strip()) if len(splts) == 3 else {}

        svc = await self._getCell(ctor, conf=svcconf)

        svc.dmon.share('svc', svc)
        root = await svc.auth.getUserByName('root')
        await root.setPasswd('root')
        info = await svc.dmon.listen('tcp://127.0.0.1:0/')
        svc.dmon.test_addr = info
        host, port = info
        surl = f'tcp://root:root@127.0.0.1:{port}/svc'
        await core.nodes(f'service.add {svcname} {surl}')
        await core.nodes(f'$lib.service.wait({svcname})')

    async def _handleStormFail(self, text):
        valu = json.loads(text)
        assert valu in (True, False), f'storm-fail must be a boolean: {text}'
        self.context['storm-fail'] = valu

    def _getStormMultiline(self, text):
        if '=' in text:
            sentinel, key = text.split('=', 1)
            if sentinel != 'MULTILINE':
                return text
            ret = self.context.get('multiline', {}).get(key)
            assert ret is not None, f'Invalid multiline text: {text}'
            return ret
        return text

    async def _handleStormMultiline(self, text):
        key, valu = text.split('=', 1)
        assert key.isupper()
        valu = json.loads(valu)
        assert isinstance(valu, str)
        multi = self.context.get('multiline', {})
        multi[key] = valu
        self.context['multiline'] = multi

    async def _handleStormOpts(self, text):
        '''
        Opts to use in subsequent Storm queries.

        Args:
            text (str): JSON string, e.g. {"vars": {"foo": "bar"}}
        '''
        item = json.loads(text)
        self.context['storm-opts'] = item

    async def _handleStormClearHttp(self, text):
        '''
        Reset the storm http context and any associated opts with it.

        Args:
            text (str): true if you also want to clear any storm/vcr opts as well
        '''
        if text == 'true':
            self.context.pop('storm-opts', None)
            self.context.pop('storm-vcr-opts', None)
            self.context.pop('storm-vcr-callback', None)
        self.context.pop('mock-http-path', None)
        self.context.pop('mock-http', None)

    async def _handleStormCortex(self, text):
        '''
        Spin up a default Cortex if ctor=default, else load the defined ctor.

        TODO: Handle providing conf in text

        Args:
            text (str): "default" or a ctor (e.g. synapse.cortex.Cortex)
        '''
        if self.core is not None:
            await self.core.fini()
            self.core = None

        ctor = 'synapse.cortex.Cortex' if text == 'default' else text

        self.core = await self._getCell(ctor)

    async def _handleStormEnvVar(self, text):
        name, valu = text.split('=', 1)
        name = name.strip()
        valu = valu.strip()
        self.stormvars[name] = os.getenv(name, valu)

    async def _handleStormExpect(self, text):
        # TODO handle some light weight output confirmation.
        return

    async def _handleStormMockHttp(self, text):
        '''
        Setup an HTTP mock file to be used with a later Storm command.

        Response file format:
        {
            "code": int,
            "body": {
                "data": json or a json str
            }
        }

        Args:
            text (str): Path to a json file with the response.
        '''
        self.context['mock-http-path'] = text

    async def _handleStormVcrOpts(self, text):
        '''
        Opts to pass to VCRPY for use in generating docs

        Args:
            text (str): JSON string, e.g. {"filter_query_args": true}
        '''
        item = json.loads(text)
        self.context['storm-vcr-opts'] = item

    async def _handleStormVcrCallback(self, text):
        '''
        Get a callback function as a dynlocal
        '''
        cb = s_dyndeps.getDynLocal(text)
        if cb is None:
            raise s_exc.NoSuchCtor(mesg=f'Failed to get callback "{text}"', ctor=text)
        self.context['storm-vcr-callback'] = cb

    async def _readline(self, line):

        match = re_directive.match(line)

        if match is not None:
            directive, text = match.groups()
            text = text.strip()

            handler = self._getHandler(directive)
            logger.debug(f'Executing {directive} -> {text}')
            await handler(text)

            return

        self._printf(line)

    async def run(self):
        '''
        Parses the specified RST file with Storm directive handling.

        Returns:
            list: List of line strings for the RST output
        '''
        with open(self.rstfile, 'r') as fd:
            lines = fd.readlines()

        for line in lines:
            await self._readline(line)

        return self.linesout





import synapse.common as s_common
import synapse.lib.crypto.coin as s_coin

import synapse.tests.utils as s_t_utils

class CryptoCoinTest(s_t_utils.SynTest):
    def test_eip55(self):
        # Test bad input on eip55
        v = s_common.guid() + 'X'
        self.none(s_coin.ether_eip55(v))



import gc
import random
import asyncio
import contextlib
import collections

import synapse.exc as s_exc
import synapse.common as s_common

import synapse.lib.coro as s_coro

from synapse.tests.utils import alist
import synapse.tests.utils as s_t_utils

class SnapTest(s_t_utils.SynTest):

    async def test_snap_eval_storm(self):

        async with self.getTestCore() as core:

            async with await core.snap() as snap:

                await snap.addNode('test:str', 'hehe')
                await snap.addNode('test:str', 'haha')

                self.len(2, await alist(snap.eval('test:str')))

                await snap.addNode('test:str', 'hoho')

                self.len(3, await alist(snap.storm('test:str')))

    async def test_snap_fires(self):
        events = []
        async with self.getTestCore() as core:

            async with await core.snap() as snap:
                def on(evnt):
                    events.append(evnt)
                snap.on('print', on)
                await snap.printf('beep')
                await snap.printf('boop')
                self.len(2, events)
                self.eq(['beep', 'boop'],
                        [m[1].get('mesg') for m in events])

                events.clear()
                snap.on('warn', on)
                await snap.warnonce('warnonce', False, key='valu')
                await snap.warnonce('warnonce', False, key='valu')
                self.len(1, events)
                self.eq(events[0], ('warn', {'mesg': 'warnonce', 'key': 'valu'}))

                events.clear()
                await snap.warn('warn', False, key='valu')
                await snap.warn('warn2', False)
                self.len(2, events)
                self.eq(events[0], ('warn', {'mesg': 'warn', 'key': 'valu'}))
                self.eq(events[1], ('warn', {'mesg': 'warn2'}))

    async def test_same_node_different_object(self):
        '''
        Test the problem in which a live node might be evicted out of the snap's buidcache causing two node
        objects to be representing the same logical thing.

        Also tests that creating a node then querying it back returns the same object
        '''
        async with self.getTestCore() as core:
            async with await core.snap() as snap:
                nodebuid = None
                snap.buidcache = collections.deque(maxlen=10)

                async def doit():
                    nonlocal nodebuid
                    # Reduce the buid cache so we don't have to make 100K nodes

                    node0 = await snap.addNode('test:int', 0)

                    node = await snap.getNodeByNdef(('test:int', 0))

                    # Test write then read coherency

                    self.eq(node0.buid, node.buid)
                    self.eq(id(node0), id(node))
                    nodebuid = node.buid

                    # Test read, then a bunch of reads, then read coherency

                    await alist(snap.addNodes([(('test:int', x), {}) for x in range(1, 20)]))
                    nodes = await alist(snap.nodesByProp('test:int'))

                    self.eq(nodes[0].buid, node0.buid)
                    self.eq(id(nodes[0]), id(node0))
                    node._test = True

                await doit()  # run in separate function so that objects are gc'd

                gc.collect()

                # Test that coherency goes away (and we don't store all nodes forever)
                await alist(snap.addNodes([(('test:int', x), {}) for x in range(20, 30)]))

                node = await snap.getNodeByNdef(('test:int', 0))
                self.eq(nodebuid, node.buid)
                # Ensure that the node is not the same object as we encountered earlier.
                # We cannot check via id() since it is possible for a pyobject to be
                # allocated at the same location as the old object.
                self.false(hasattr(node, '_test'))

    async def test_addNodes(self):
        async with self.getTestCore() as core:
            async with await core.snap() as snap:
                ndefs = ()
                self.len(0, await alist(snap.addNodes(ndefs)))

                ndefs = (
                    (('test:str', 'hehe'), {'props': {'.created': 5, 'tick': 3}, 'tags': {'cool': (1, 2)}}, ),
                )
                result = await alist(snap.addNodes(ndefs))
                self.len(1, result)

                node = result[0]
                self.eq(node.props.get('tick'), 3)
                self.ge(node.props.get('.created', 0), 5)
                self.eq(node.tags.get('cool'), (1, 2))

                nodes = await alist(snap.nodesByPropValu('test:str', '=', 'hehe'))
                self.len(1, nodes)
                self.eq(nodes[0], node)

                # Make sure that we can still add secondary props even if the node already exists
                node2 = await snap.addNode('test:str', 'hehe', props={'baz': 'test:guid:tick=2020'})
                self.eq(node2, node)
                self.nn(node2.get('baz'))

    async def test_addNodesAuto(self):
        '''
        Secondary props that are forms when set make nodes
        '''
        async with self.getTestCore() as core:
            async with await core.snap() as snap:

                node = await snap.addNode('test:guid', '*')
                await node.set('size', 42)
                nodes = await alist(snap.nodesByPropValu('test:int', '=', 42))
                self.len(1, nodes)

                # For good measure, set a secondary prop that is itself a comp type that has an element that
                # is a form
                node = await snap.addNode('test:haspivcomp', 42)
                await node.set('have', ('woot', 'rofl'))
                nodes = await alist(snap.nodesByPropValu('test:pivcomp', '=', ('woot', 'rofl')))
                self.len(1, nodes)
                nodes = await alist(snap.nodesByProp('test:pivcomp:lulz'))
                self.len(1, nodes)
                nodes = await alist(snap.nodesByPropValu('test:str', '=', 'rofl'))
                self.len(1, nodes)

                # Make sure the sodes didn't get misordered
                node = await snap.addNode('inet:dns:a', ('woot.com', '1.2.3.4'))
                self.eq(node.ndef[0], 'inet:dns:a')

    async def test_addNodeRace(self):
        ''' Test when a reader might retrieve a partially constructed node '''
        NUM_TASKS = 2
        failed = False
        done_events = []

        data = list(range(50))
        rnd = random.Random()
        rnd.seed(4)  # chosen by fair dice roll
        rnd.shuffle(data)

        async with self.getTestCore() as core:

            async def write_a_bunch(done_event):
                nonlocal failed
                await asyncio.sleep(0)
                async with await core.snap() as snap:

                    async def waitabit(info):
                        await asyncio.sleep(0.1)

                    snap.on('node:add', waitabit)

                    for i in data:
                        node = await snap.addNode('test:int', i)
                        if node.props.get('.created') is None:
                            failed = True
                            done_event.set()
                            return
                        await asyncio.sleep(0)
                done_event.set()

            for _ in range(NUM_TASKS):
                done_event = asyncio.Event()
                core.schedCoro(write_a_bunch(done_event))
                done_events.append(done_event)

            for event in done_events:
                await event.wait()

            self.false(failed)

    async def test_addNodeRace2(self):
        ''' Test that dependencies between active editatoms don't wedge '''
        bc_done_event = asyncio.Event()
        ab_middle_event = asyncio.Event()
        ab_done_event = asyncio.Event()

        async with self.getTestCore() as core:
            async def bc_writer():
                async with await core.snap() as snap:
                    call_count = 0

                    origGetNodeByBuid = snap.getNodeByBuid

                    async def slowGetNodeByBuid(buid):
                        nonlocal call_count
                        call_count += 1
                        if call_count > 0:
                            await ab_middle_event.wait()
                        return await origGetNodeByBuid(buid)

                    snap.getNodeByBuid = slowGetNodeByBuid

                    await snap.addNode('test:pivcomp', ('woot', 'rofl'))
                bc_done_event.set()

            core.schedCoro(bc_writer())
            await asyncio.sleep(0)

            async def ab_writer():
                async with await core.snap() as snap:

                    origGetNodeByBuid = snap.getNodeByBuid

                    async def slowGetNodeByBuid(buid):
                        ab_middle_event.set()
                        return await origGetNodeByBuid(buid)

                    snap.getNodeByBuid = slowGetNodeByBuid

                    await snap.addNode('test:haspivcomp', 42, props={'have': ('woot', 'rofl')})
                ab_done_event.set()

            core.schedCoro(ab_writer())
            self.true(await s_coro.event_wait(bc_done_event, 5))
            self.true(await s_coro.event_wait(ab_done_event, 5))

    @contextlib.asynccontextmanager
    async def _getTestCoreMultiLayer(self):
        '''
        Create a cortex with a second view which has an additional layer above the main layer.

        Notes:
            This method is broken out so subclasses can override.
        '''
        async with self.getTestCore() as core0:

            view0 = core0.view
            layr0 = view0.layers[0]

            ldef1 = await core0.addLayer()
            layr1 = core0.getLayer(ldef1.get('iden'))
            vdef1 = await core0.addView({'layers': [layr1.iden, layr0.iden]})

            yield view0, core0.getView(vdef1.get('iden'))

    async def test_cortex_lift_layers_simple(self):
        async with self._getTestCoreMultiLayer() as (view0, view1):
            ''' Test that you can write to view0 and read it from view1 '''
            self.len(1, await alist(view0.eval('[ inet:ipv4=1.2.3.4 :asn=42 +#woot=(2014, 2015)]')))
            self.len(1, await alist(view1.eval('inet:ipv4')))
            self.len(1, await alist(view1.eval('inet:ipv4=1.2.3.4')))
            self.len(1, await alist(view1.eval('inet:ipv4:asn=42')))
            self.len(1, await alist(view1.eval('inet:ipv4 +:asn=42')))
            self.len(1, await alist(view1.eval('inet:ipv4 +#woot')))

    async def test_cortex_lift_layers_bad_filter(self):
        '''
        Test a two layer cortex where a lift operation gives the wrong result
        '''
        async with self._getTestCoreMultiLayer() as (view0, view1):

            self.len(1, await alist(view0.eval('[ inet:ipv4=1.2.3.4 :asn=42 +#woot=(2014, 2015)]')))
            self.len(1, await alist(view1.eval('inet:ipv4#woot@=2014')))
            self.len(1, await alist(view1.eval('inet:ipv4=1.2.3.4 [ :asn=31337 +#woot=2016 ]')))

            self.len(0, await alist(view0.eval('inet:ipv4:asn=31337')))
            self.len(1, await alist(view1.eval('inet:ipv4:asn=31337')))

            self.len(1, await alist(view0.eval('inet:ipv4:asn=42')))
            self.len(0, await alist(view1.eval('inet:ipv4:asn=42')))

            self.len(1, await alist(view0.eval('[ test:arrayprop="*" :ints=(1, 2, 3) ]')))
            self.len(1, await alist(view1.eval('test:int=2 -> test:arrayprop')))
            self.len(1, await alist(view1.eval('test:arrayprop [ :ints=(4, 5, 6) ]')))

            self.len(0, await alist(view0.eval('test:int=5 -> test:arrayprop')))
            self.len(1, await alist(view1.eval('test:int=5 -> test:arrayprop')))

            self.len(1, await alist(view0.eval('test:int=2 -> test:arrayprop')))
            self.len(0, await alist(view1.eval('test:int=2 -> test:arrayprop')))

            self.len(1, await alist(view1.eval('[ test:int=7 +#atag=2020 ]')))
            self.len(1, await alist(view0.eval('[ test:int=7 +#atag=2021 ]')))

            self.len(0, await alist(view0.eval('test:int#atag@=2020')))
            self.len(1, await alist(view1.eval('test:int#atag@=2020')))

            self.len(1, await alist(view0.eval('test:int#atag@=2021')))
            self.len(0, await alist(view1.eval('test:int#atag@=2021')))

    async def test_cortex_lift_layers_dup(self):
        '''
        Test a two layer cortex where a lift operation might give the same node twice incorrectly
        '''
        async with self._getTestCoreMultiLayer() as (view0, view1):
            # add to view1 first so we can cause creation in both...
            self.len(1, await alist(view1.eval('[ inet:ipv4=1.2.3.4 :asn=42 ]')))
            self.len(1, await alist(view0.eval('[ inet:ipv4=1.2.3.4 :asn=42 ]')))

            # lift by primary and ensure only one...
            self.len(1, await alist(view1.eval('inet:ipv4')))

            # lift by secondary and ensure only one...
            self.len(1, await alist(view1.eval('inet:ipv4:asn=42')))

            # now set one to a diff value that we will ask for but should be masked
            self.len(1, await alist(view0.eval('[ inet:ipv4=1.2.3.4 :asn=99 ]')))
            self.len(0, await alist(view1.eval('inet:ipv4:asn=99')))

    async def test_cortex_lift_bytype(self):
        async with self.getTestCore() as core:
            await core.nodes('[ inet:dns:a=(vertex.link, 1.2.3.4) ]')
            nodes = await core.nodes('inet:ipv4*type=1.2.3.4')
            self.len(2, nodes)
            self.eq(nodes[0].ndef, ('inet:ipv4', 0x01020304))
            self.eq(nodes[1].ndef, ('inet:dns:a', ('vertex.link', 0x01020304)))

    async def test_clearcache(self):

        # Type hinting since we dont do the type hinting
        # properly in the Cortex anymore... :(
        import synapse.lib.snap as s_snap

        async with self.getTestCore() as core:
            async with await core.snap() as snap0:  # type: s_snap.Snap

                original_node0 = await snap0.addNode('test:str', 'node0')
                self.len(1, snap0.buidcache)
                self.len(1, snap0.livenodes)
                self.len(0, snap0.tagcache)
                self.len(0, snap0.tagnorms)

                await original_node0.addTag('foo.bar.baz')
                self.len(4, snap0.buidcache)
                self.len(4, snap0.livenodes)
                self.len(3, snap0.tagnorms)

                async with await core.snap() as snap1:  # type: s_snap.Snap
                    snap1_node0 = await snap1.getNodeByNdef(('test:str', 'node0'))
                    await snap1_node0.delTag('foo.bar.baz')
                    self.notin('foo.bar.baz', snap1_node0.tags)
                    # Our reference to original_node0 still has the tag though
                    self.isin('foo.bar.baz', original_node0.tags)

                # We rely on the layer's row cache to be correct in this test.

                # Lift is cached..
                same_node0 = await snap0.getNodeByNdef(('test:str', 'node0'))
                self.eq(id(original_node0), id(same_node0))

                # flush snap0 cache!
                await snap0.clearCache()
                self.len(0, snap0.buidcache)
                self.len(0, snap0.livenodes)
                self.len(0, snap0.tagcache)
                self.len(0, snap0.tagnorms)

                # After clearing the cache and lifting nodes, the new node
                # was lifted directly from the layer.
                new_node0 = await snap0.getNodeByNdef(('test:str', 'node0'))
                self.ne(id(original_node0), id(new_node0))
                self.notin('foo.bar.baz', new_node0.tags)

    async def test_cortex_lift_layers_bad_filter_tagprop(self):
        '''
        Test a two layer cortex where a lift operation gives the wrong result, with tagprops
        '''
        async with self._getTestCoreMultiLayer() as (view0, view1):
            await view0.core.addTagProp('score', ('int', {}), {'doc': 'hi there'})

            self.len(1, await view0.nodes('[ test:int=10 +#woot:score=20 ]'))
            self.len(1, await view1.nodes('#woot:score=20'))
            self.len(1, await view1.nodes('[ test:int=10 +#woot:score=40 ]'))

            self.len(0, await view0.nodes('#woot:score=40'))
            self.len(1, await view1.nodes('#woot:score=40'))

            self.len(1, await view0.nodes('#woot:score=20'))
            self.len(0, await view1.nodes('#woot:score=20'))

    async def test_cortex_lift_layers_dup_tagprop(self):
        '''
        Test a two layer cortex where a lift operation might give the same node twice incorrectly
        '''
        async with self._getTestCoreMultiLayer() as (view0, view1):
            await view0.core.addTagProp('score', ('int', {}), {'doc': 'hi there'})

            self.len(1, await view1.nodes('[ test:int=10 +#woot:score=20 ]'))
            self.len(1, await view0.nodes('[ test:int=10 +#woot:score=20 ]'))

            self.len(1, await view1.nodes('#woot:score=20'))

            self.len(1, await view0.nodes('[ test:int=10 +#woot:score=40 ]'))

    async def test_cortex_lift_layers_ordering(self):

        async with self._getTestCoreMultiLayer() as (view0, view1):

            await view0.core.addTagProp('score', ('int', {}), {'doc': 'hi there'})
            await view0.core.addTagProp('data', ('data', {}), {'doc': 'hi there'})

            await view0.nodes('[ inet:ipv4=1.1.1.4 ]')
            await view1.nodes('inet:ipv4=1.1.1.4 [+#tag]')
            await view0.nodes('inet:ipv4=1.1.1.4 | delnode')
            nodes = await view1.nodes('#tag | uniq')
            self.len(0, nodes)

            await view0.nodes('[ inet:ipv4=1.1.1.4 :asn=4 +#woot:score=4] $node.data.set(woot, 4)')
            await view0.nodes('[ inet:ipv4=1.1.1.1 :asn=1 +#woot:score=1] $node.data.set(woot, 1)')
            await view1.nodes('[ inet:ipv4=1.1.1.2 :asn=2 +#woot:score=2] $node.data.set(woot, 2)')
            await view0.nodes('[ inet:ipv4=1.1.1.3 :asn=3 +#woot:score=3] $node.data.set(woot, 3)')

            await view1.nodes('[ test:str=foo +#woot=2001 ]')
            await view0.nodes('[ test:str=foo +#woot=2001 ]')
            await view0.nodes('[ test:int=1 +#woot=2001 ]')
            await view0.nodes('[ test:int=2 +#woot=2001 ]')

            nodes = await view1.nodes('#woot')
            self.len(7, nodes)

            nodes = await view1.nodes('inet:ipv4')
            self.len(4, nodes)
            last = 0
            for node in nodes:
                valu = node.ndef[1]
                self.gt(valu, last)
                last = valu

            nodes = await view1.nodes('inet:ipv4:asn')
            self.len(4, nodes)
            last = 0
            for node in nodes:
                asn = node.props.get('asn')
                self.gt(asn, last)
                last = asn

            nodes = await view1.nodes('inet:ipv4:asn>0')
            self.len(4, nodes)
            last = 0
            for node in nodes:
                asn = node.props.get('asn')
                self.gt(asn, last)
                last = asn

            nodes = await view1.nodes('inet:ipv4:asn*in=(1,2,3,4)')
            self.len(4, nodes)
            last = 0
            for node in nodes:
                asn = node.props.get('asn')
                self.gt(asn, last)
                last = asn

            nodes = await view1.nodes('inet:ipv4:asn*in=(4,3,2,1)')
            self.len(4, nodes)
            last = 5
            for node in nodes:
                asn = node.props.get('asn')
                self.lt(asn, last)
                last = asn

            nodes = await view1.nodes('#woot:score')
            self.len(4, nodes)
            last = 0
            for node in nodes:
                scor = node.getTagProp('woot', 'score')
                self.gt(scor, last)
                last = scor

            nodes = await view1.nodes('#woot:score>0')
            self.len(4, nodes)
            last = 0
            for node in nodes:
                scor = node.getTagProp('woot', 'score')
                self.gt(scor, last)
                last = scor

            nodes = await view1.nodes('#woot:score*in=(1,2,3,4)')
            self.len(4, nodes)
            last = 0
            for node in nodes:
                scor = node.getTagProp('woot', 'score')
                self.gt(scor, last)
                last = scor

            nodes = await view1.nodes('#woot:score*in=(4,3,2,1)')
            self.len(4, nodes)
            last = 5
            for node in nodes:
                scor = node.getTagProp('woot', 'score')
                self.lt(scor, last)
                last = scor

            await view0.nodes('[ test:arrayform=(3,5,6)]')
            await view0.nodes('[ test:arrayform=(1,2,3)]')
            await view1.nodes('[ test:arrayform=(2,3,4)]')
            await view0.nodes('[ test:arrayform=(3,4,5)]')

            nodes = await view1.nodes('test:arrayform*[=3]')
            self.len(4, nodes)

            nodes = await view1.nodes('test:arrayform*[=2]')
            self.len(2, nodes)

            nodes = await view1.nodes('yield $lib.lift.byNodeData(woot)')
            self.len(4, nodes)

            self.len(1, await view1.nodes('[crypto:x509:cert="*" :identities:fqdns=(somedomain.biz,www.somedomain.biz)]'))
            nodes = await view1.nodes('crypto:x509:cert:identities:fqdns*[="*.biz"]')
            self.len(2, nodes)

            self.len(1, await view1.nodes('[crypto:x509:cert="*" :identities:fqdns=(somedomain.biz,www.somedomain.biz)]'))
            nodes = await view1.nodes('crypto:x509:cert:identities:fqdns*[="*.biz"]')
            self.len(4, nodes)

            await view0.nodes('[ test:data=(123) :data=(123) +#woot:data=(123)]')
            await view1.nodes('[ test:data=foo :data=foo +#woot:data=foo]')
            await view0.nodes('[ test:data=(0) :data=(0) +#woot:data=(0)]')
            await view0.nodes('[ test:data=bar :data=foo +#woot:data=foo]')

            nodes = await view1.nodes('test:data')
            self.len(4, nodes)

            nodes = await view1.nodes('test:data=foo')
            self.len(1, nodes)

            nodes = await view1.nodes('test:data:data')
            self.len(4, nodes)

            nodes = await view1.nodes('test:data:data=foo')
            self.len(2, nodes)

            nodes = await view1.nodes('#woot:data')
            self.len(4, nodes)

            nodes = await view1.nodes('#woot:data=foo')
            self.len(2, nodes)

    async def test_snap_editor(self):

        async with self.getTestCore() as core:
            nodes = await core.nodes('[ media:news=63381924986159aff183f0c85bd8ebad +(refs)> {[ inet:fqdn=vertex.link ]} ]')
            root = core.auth.rootuser
            async with await core.view.snap(user=root) as snap:
                async with snap.getEditor() as editor:
                    fqdn = await editor.addNode('inet:fqdn', 'vertex.link')
                    news = await editor.addNode('media:news', '63381924986159aff183f0c85bd8ebad')
                    self.false(await news.addEdge('refs', s_common.ehex(fqdn.buid)))
                    self.len(0, editor.getNodeEdits())

                    self.true(await news.addEdge('pwns', s_common.ehex(fqdn.buid)))
                    nodeedits = editor.getNodeEdits()
                    self.len(1, nodeedits)
                    self.len(1, nodeedits[0][2])

            self.len(1, await core.nodes('media:news -(pwns)> *'))

            self.len(1, await core.nodes('[ inet:dns:soa=(lol,) :fqdn=vertex.link ]'))
            self.len(1, await core.nodes('inet:dns:soa=(lol,) [ :fqdn=vertex.link ]'))
            with self.raises(s_exc.ReadOnlyProp):
                await core.nodes('inet:dns:soa=(lol,) [ :fqdn=woot.com ]')



import logging

import synapse.common as s_common

import synapse.lib.config as s_config

logger = logging.getLogger(__name__)

rstlvls = [
    ('#', {'over': True}),
    ('*', {'over': True}),
    ('=', {}),
    ('-', {}),
    ('^', {}),
]

stormtype_doc_schema = {
    'definitions': {

        'stormType': {
            'type': ['string', 'array', 'object'],
            'items': {'type': 'string'},
            'properties': {
                'name': {'type': 'string',
                         'description': 'For a function argument, the name of the argument.'},
                'desc': {'type': 'string',
                         'description': 'For a function argument or return value, the description of the value.'},
                'type': {'$ref': '#/definitions/stormType'},
                'args': {
                    'type': 'array',
                    'items': {'$ref': '#/definitions/stormType'},
                    'description': 'Arguments to document.',
                },
                'returns': {'$ref': '#/definitions/stormType',
                            'description': 'Function return types to document'},
                'default': {'type': ['boolean', 'integer', 'string', 'null'],
                            'description': 'For a function argument, the default value, if applicable.'},
            },
            'required': ['type'],
            'description': 'A multi-purpose container for holding types information. If this '
                           'is a string or list of strings, it represents simple return types.'
                           ' If it is a object, it should represent a function to generate '
                           'documentation for.',
            'additionalProperties': False,
        },

        'stormtypeDoc': {
            'type': 'object',
            'properties': {
                'name': {'type': 'string',
                         'description': 'The name of the object.'},
                'desc': {'type': 'string',
                         'description': 'The docstring of the object.'},
                'type': {'$ref': '#/definitions/stormType'}
            },
            'additionalProperties': False,
        },

    },
    'type': 'object',
    'properties': {
        'path': {
            'type': 'array',
            'items': {
                'type': 'string'
            },
            'minItems': 1,
            'description': 'The path of the object.'
        },
        'desc': {
            'type': 'string',
            'description': 'The doc for the object itself.'
        },
        'locals': {
            'type': 'array',
            'items': {'$ref': '#/definitions/stormtypeDoc'},
            'description': 'A list of attributes or functions to document.',
        },
        'additionalProperties': False,
    },
}
reqValidStormTypeDoc = s_config.getJsValidator(stormtype_doc_schema)

class RstHelp:

    def __init__(self):
        self.lines = []

    def addHead(self, name, lvl=0, link=None):
        char, info = rstlvls[lvl]
        under = char * len(name)

        lines = []

        lines.append('')

        if link:
            lines.append('')
            lines.append(link)
            lines.append('')

        if info.get('over'):
            lines.append(under)

        lines.append(name)
        lines.append(under)
        lines.append('')

        self.addLines(*lines)

    def addLines(self, *lines):
        self.lines.extend(lines)

    def getRstText(self):
        return '\n'.join(self.lines)

def ljuster(ilines):
    '''Helper to lstrip lines of whitespace an appropriate amount.'''
    baseline = ilines[0]
    assert baseline != ''
    newbaseline = baseline.lstrip()
    assert newbaseline != ''
    diff = len(baseline) - len(newbaseline)
    assert diff >= 0
    newlines = [line[diff:] for line in ilines]
    return newlines

def scrubLines(lines):
    '''Remove any empty lines until we encounter non-empty linee'''
    newlines = []
    for line in lines:
        if line == '' and not newlines:
            continue
        newlines.append(line)

    return newlines

def prepareRstLines(doc):
    '''Prepare a desc string for RST lines.'''
    lines = doc.split('\n')
    lines = scrubLines(lines)
    lines = ljuster(lines)
    return lines

def getArgLines(rtype):
    lines = []
    args = rtype.get('args', ())
    assert args is not None

    if args == ():
        # Zero args
        return lines

    lines.append('\n')
    lines.append('Args:')
    for arg in args:
        name = arg.get('name')
        desc = arg.get('desc')
        atyp = arg.get('type')
        assert name is not None
        assert desc is not None
        assert atyp is not None
        name = name.replace('*', '\\*')
        if isinstance(atyp, str):
            line = f'    {name} ({atyp}): {desc}'
        elif isinstance(atyp, (list, tuple)):
            assert len(atyp) > 1
            for obj in atyp:
                assert isinstance(obj, str)
            tdata = ', '.join([f'``{obj}``' for obj in atyp])
            rline = f'The input type may one one of the following: {tdata}.'
            line = f'    {name}: {desc} {rline}'
        elif isinstance(atyp, dict):
            logger.warning('Fully declarative input types are not yet supported.')
            rline = f"The input type is derived from the declarative type ``{atyp}``."
            line = f'    {name}: {desc} {rline}'
        else:
            raise AssertionError(f'unknown argtype: {atyp}')

        lines.extend((line, '\n'))

    return lines

def genCallsig(rtype):
    items = []

    args = rtype.get('args', ())
    assert args is not None
    for arg in args:
        name = arg.get('name')
        defv = arg.get('default', s_common.novalu)

        item = name
        if defv is not s_common.novalu:
            item = f'{item}={defv}'

        items.append(item)

    ret = f"({', '.join(items)})"
    return ret

def getLink(sname, linkprefix, ref=False, suffix=None):
    sname = sname.replace(":", ".").replace(".", "-")
    if suffix:
        sname = f'{sname}-{suffix}'
    if ref:
        link = f':ref:`{linkprefix}-{sname}`'
    else:
        link = f'.. _{linkprefix}-{sname}:'
    return link

def getRtypeStr(rtype, known_types, types_prefix, suffix):
    if rtype in known_types:
        rtype = getLink(rtype, types_prefix, ref=True, suffix=suffix)
    else:
        rtype = f'``{rtype}``'
    return rtype

def getReturnLines(rtype, known_types=None, types_prefix=None, suffix=None):
    # Allow someone to plumb in name=Yields as a return type.
    lines = ['']
    if known_types is None:
        known_types = set()

    if isinstance(rtype, str):
        lines.append('Returns:')
        lines.append(f'    The type is {getRtypeStr(rtype, known_types, types_prefix, suffix)}.')
    elif isinstance(rtype, (list, tuple)):
        assert len(rtype) > 1
        tdata = ', '.join([f'{getRtypeStr(obj, known_types, types_prefix, suffix)}' for obj in rtype])
        lines.append('Returns:')
        lines.append(f'    The type may be one of the following: {tdata}.')
    elif isinstance(rtype, dict):
        returns = rtype.get('returns')
        assert returns is not None, f'Invalid returns for {rtype}'
        name = returns.get('name', 'Returns')

        desc = returns.get('desc')
        rettype = returns.get('type')

        lines.append(f'{name}:')
        # Now switch on the type.

        parts = ['   ']
        if desc:
            parts.append(desc)

        if isinstance(rettype, str):
            parts.append(f"The return type is {getRtypeStr(rettype, known_types, types_prefix, suffix)}.")
        elif isinstance(rettype, (list, tuple)):
            assert len(rettype) > 1
            tdata = ', '.join([f'{getRtypeStr(obj, known_types, types_prefix, suffix)}' for obj in rettype])
            rline = f'The return type may be one of the following: {tdata}.'
            parts.append(rline)
        elif isinstance(rettype, dict):
            logger.warning('Fully declarative input types are not yet supported.')
            rline = f"The return type is derived from the declarative type ``{rettype}``."
            parts.append(rline)
        else:
            raise AssertionError(f'unknown return type: {rettype}')
        line = ' '.join(parts)
        lines.append(line)
    return lines

def docStormTypes(page, docinfo, linkprefix, islib=False, lvl=1,
                  known_types=None, types_prefix=None, types_suffix=None,
                  ):
    '''
    Process a list of StormTypes doc information to add them to a a RstHelp object.

    Notes
        This will create internal hyperlink link targets for each header item. The
        link prefix string must be given with the ``linkprefix`` argument.

    Args:
        page (RstHelp): The RST page to add .
        docinfo (dict): A Stormtypes Doc.
        linkprefix (str): The RST link prefix string to use.
        islib (bool): Treat the data as a library. This will preface the header and
            attribute values with ``$`` and use full paths for attributes.
        lvl (int): The base header level to use when adding headers to the page.

    Returns:
        None
    '''

    if known_types is None:
        known_types = set()

    for info in docinfo:
        reqValidStormTypeDoc(info)

        path = info.get('path')

        sname = '.'.join(path)

        safesname = sname.replace(':', '\\:')
        if islib:
            link = getLink(sname, linkprefix)
            page.addHead(f"${safesname}", lvl=lvl, link=link)
        else:
            link = getLink(sname, linkprefix, suffix=types_suffix)
            page.addHead(safesname, lvl=lvl, link=link)

        typedoc = info.get('desc')
        lines = prepareRstLines(typedoc)

        page.addLines(*lines)

        locls = info.get('locals')
        locls = sorted(locls, key=lambda x: x.get('name'))

        for locl in locls:

            name = locl.get('name')
            loclname = '.'.join((sname, name))
            desc = locl.get('desc')
            rtype = locl.get('type')
            assert desc is not None
            assert rtype is not None

            link = f'.. _{linkprefix}-{loclname.replace(":", ".").replace(".", "-")}:'

            if isinstance(rtype, dict):
                rname = rtype.get('type')
                assert rname == 'function', f'Unknown type: loclname={loclname} rname={rname}'

                lines = prepareRstLines(desc)
                arglines = getArgLines(rtype)
                lines.extend(arglines)

                retlines = getReturnLines(rtype, known_types=known_types, types_prefix=types_prefix,
                                          suffix=types_suffix)
                lines.extend(retlines)

                callsig = genCallsig(rtype)
                header = f'{name}{callsig}'
                header = header.replace('*', r'\*')

            else:
                header = name
                lines = prepareRstLines(desc)

                retlines = getReturnLines(rtype, known_types=known_types, types_prefix=types_prefix,
                                          suffix=types_suffix)
                lines.extend(retlines)

            if islib:
                header = '.'.join((safesname, header))
                header = f'${header}'

            page.addHead(header, lvl=lvl + 1, link=link)
            page.addLines(*lines)



import synapse.lib.module as s_module

class GovUsModule(s_module.CoreModule):

    def getModelDefs(self):
        modl = {
            'types': (
                ('gov:us:ssn', ('int', {}), {'doc': 'A US Social Security Number (SSN).'}),
                ('gov:us:zip', ('int', {}), {'doc': 'A US Postal Zip Code.'}),
                ('gov:us:cage', ('str', {'lower': True}), {'doc': 'A Commercial and Government Entity (CAGE) code.'}),
            ),

            'forms': (
                ('gov:us:cage', {}, (
                    ('name0', ('ou:name', {}), {'doc': 'The name of the organization.'}),
                    ('name1', ('str', {'lower': True}), {'doc': 'Name Part 1.'}),
                    ('street', ('str', {'lower': True}), {}),
                    ('city', ('str', {'lower': True}), {}),
                    ('state', ('str', {'lower': True}), {}),
                    ('zip', ('gov:us:zip', {}), {}),
                    ('cc', ('pol:iso2', {}), {}),
                    ('country', ('str', {'lower': True}), {}),
                    ('phone0', ('tel:phone', {}), {}),
                    ('phone1', ('tel:phone', {}), {}),
                )),

                ('gov:us:ssn', {}, []),
                ('gov:us:zip', {}, []),
            ),
        }
        name = 'gov:us'
        return ((name, modl), )




import synapse.exc as s_exc
import synapse.common as s_common

import synapse.lib.stormtypes as s_stormtypes

@s_stormtypes.registry.registerLib
class LibWhois(s_stormtypes.Lib):
    '''
    A Storm Library for providing a consistent way to generate guids for WHOIS / Registration Data in Storm.
    '''
    _storm_locals = (
        {'name': 'guid',
         'desc': '''
        Provides standard patterns for creating guids for certain inet:whois forms.

        Raises:
            StormRuntimeError: If form is not supported in this method.''',
         'type': {'type': 'function', '_funcname': '_whoisGuid',
                  'args': (
                      {'name': 'props', 'type': 'dict', 'desc': 'Dictionary of properties used to create the form.', },
                      {'name': 'form', 'type': 'str', 'desc': 'The ``inet:whois`` form to create the guid for.', },
                  ),
                  'returns': {'type': 'str', 'desc': 'A guid for creating a the node for.', }}},
    )
    _storm_lib_path = ('inet', 'whois')

    def getObjLocals(self):
        return {
             'guid': self._whoisGuid,
         }

    async def _whoisGuid(self, props, form):
        form = await s_stormtypes.tostr(form)
        props = await s_stormtypes.toprim(props)
        if form == 'iprec':
            guid_props = ('net4', 'net6', 'asof', 'id')
        elif form == 'ipcontact':
            guid_props = ('contact', 'asof', 'id', 'updated')
        elif form == 'ipquery':
            guid_props = ('time', 'fqdn', 'url', 'ipv4', 'ipv6')
        else:
            mesg = f'No guid helpers available for this inet:whois form'
            raise s_exc.StormRuntimeError(mesg=mesg, form=form)

        guid_vals = []
        try:
            for prop in guid_props:
                val = props.get(prop)
                if val is not None:
                    guid_vals.append(str(val))
        except AttributeError as e:
            mesg = f'Failed to iterate over props {str(e)}'
            raise s_exc.StormRuntimeError(mesg=mesg)

        if len(guid_vals) <= 1:
            await self.runt.snap.warn(f'Insufficient guid vals identified, using random guid: {guid_vals}')
            return s_common.guid()

        return s_common.guid(sorted(guid_vals))



import synapse.exc as s_exc
import synapse.common as s_common

import synapse.lib.coro as s_coro
import synapse.lib.storm as s_storm
import synapse.lib.stormtypes as s_stormtypes

import xml.etree.ElementTree as xml_et

@s_stormtypes.registry.registerType
class XmlElement(s_stormtypes.Prim):
    '''
    A Storm object for dealing with elements in an XML tree.
    '''
    _storm_typename = 'xml:element'
    _storm_locals = (
        {'name': 'name', 'desc': 'The element tag name.', 'type': 'str'},
        {'name': 'text', 'desc': 'The element text body.', 'type': 'str'},
        {'name': 'attrs', 'desc': 'The element attributes list.', 'type': 'dict'},

        {'name': 'find', 'desc': 'Find all nested elements with the specified tag name.',
         'type': {'type': 'function', '_funcname': 'find',
                  'args': (
                      {'name': 'name', 'type': 'str', 'desc': 'The name of the XML tag.'},
                      {'name': 'nested', 'type': 'bool', 'default': True,
                        'desc': 'Set to $lib.false to only find direct children.'},
                  ),
                  'returns': {'type': 'generator', 'desc': 'A generator which yields xml:elements.'}}},

        {'name': 'get', 'desc': 'Get a single child element by XML tag name.',
         'type': {'type': 'function', '_funcname': 'get',
                  'args': (
                      {'name': 'name', 'type': 'str', 'desc': 'The name of the child XML element tag.'},
                  ),
                  'returns': {'type': 'xml:element', 'desc': 'The child XML element or $lib.null'}}},
    )

    def __init__(self, runt, elem):
        s_stormtypes.Prim.__init__(self, None)
        self.runt = runt
        self.elem = elem
        self.locls.update({
            'get': self.get,
            'find': self.find,
            'name': elem.tag,
            'text': elem.text,
            'attrs': elem.attrib,
        })

    async def iter(self):
        for elem in self.elem:
            yield XmlElement(self.runt, elem)

    async def find(self, name, nested=True):
        name = await s_stormtypes.tostr(name)
        nested = await s_stormtypes.tobool(nested)

        if nested:
            for elem in self.elem.iter(name):
                yield XmlElement(self.runt, elem)
        else:
            for elem in self.elem.findall(name):
                yield XmlElement(self.runt, elem)

    async def get(self, name):
        name = await s_stormtypes.tostr(name)
        elem = self.elem.find(name)
        if elem is not None:
            return XmlElement(self.runt, elem)

@s_stormtypes.registry.registerLib
class LibXml(s_stormtypes.Lib):
    '''
    A Storm library for parsing XML.
    '''
    _storm_lib_path = ('xml',)
    _storm_locals = (
        {'name': 'parse', 'desc': 'Parse an XML string into an xml:element tree.',
         'type': {'type': 'function', '_funcname': 'parse',
            'args': (
                {'name': 'valu', 'type': 'str', 'desc': 'The XML string to parse into an xml:element tree.'},
            ),
            'returns': {'type': 'xml:element', 'desc': 'An xml:element for the root node of the XML tree.'},
        }},
    )

    def getObjLocals(self):
        return {
            'parse': self.parse,
        }

    async def parse(self, valu):
        valu = await s_stormtypes.tostr(valu)
        try:
            todo = (xml_et.fromstring, (valu,), {})
            root = await s_coro.spawn(todo)
            return XmlElement(self.runt, root)
        except xml_et.ParseError as e:
            raise s_exc.BadArg(mesg=f'Invalid XML text: {str(e)}')
